# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''
Module for handling atoms and molecules.
'''

from math import sqrt
import numpy as np
from mendeleev import element


class Atom(object):

    '''Basic atom class representing an atom.'''

    def __init__(self, identifier, xyz=(0.0, 0.0, 0.0), dummy=False, id=None):

        self.xyz = np.asarray(xyz)
        self.is_dummy = dummy
        self._set_attributes(identifier)

    @property
    def xyz(self):
        return self._xyz

    @xyz.setter
    def xyz(self, values):

        if len(values) != 3:
            raise ValueError("Expecting 3 coordinates (x, y, z), got: {0:d}".format(len(values)))
        else:
            self._xyz = np.asarray(values)

    def _set_attributes(self, identifier):
        '''
        Set the attributes of an atom based on the unique "indentifier" provided.
        The attributes are read from the elements.json file.
        '''

        attrs = ["name", "symbol", "atomic_number", "mass"]
        atom = element(identifier)

        for attr in attrs:
            setattr(self, attr, getattr(atom, attr))

        if self.is_dummy:
            self.set_atomic_number(0.0)

    def set_atomic_number(self, value):
        self.atomic_number = value

    def move(self, x=0.0, y=0.0, z=0.0):

        '''Move atom to a set of new coordinates given in xyz'''

        self.xyz = np.asarray([x, y, z], dtype=self._dtxyz)

    def gamess_rep(self):

        out = "{0:<10s} {1:5.1f}\t{2:15.5f}{3:15.5f}{4:15.5f}\n".format(
                self.symbol, float(self.atomic_number), self.xyz[0], self.xyz[1], self.xyz[2])
        return out

    def __repr__(self):
        out = "{0:<10s} {1:5.1f}\t{2:15.5f}{3:15.5f}{4:15.5f}".format(
                self.symbol, float(self.atomic_number), self.xyz[0], self.xyz[1], self.xyz[2])
        return out

    def __str__(self):
        out = "{0:<10s} {1:14.2f}\t{2:15.5f}{3:15.5f}{4:15.5f}\n".format(
                self.symbol, float(self.atomic_number), self.xyz[0], self.xyz[1], self.xyz[2])
        return out


class Molecule(object):

    '''Molecule class handling all the operations on single molecules'''

    def __init__(self, name="", atoms=None, unique=None, sym="", charge=0, multiplicity=1):
        self.name = name
        self.charge = charge
        self.multiplicity = multiplicity
        self.atoms = atoms
        if sym == "":
            self.symmetry = "c1"
        else:
            self.symmetry = sym
        self.electrons = self.nele()
        if unique is not None:
            self.unique_labels = unique
        else:
            self.unique_labels = range(len(atoms))

    @property
    def atoms(self):
        return self._atoms

    @atoms.setter
    def atoms(self, values):
        self._atoms = []
        if hasattr(values, "__iter__"):
            for v in values:
                if len(v) == 1:
                    self._atoms.append(Atom(identifier=v[0]))
                elif len(v) == 2:
                    if isinstance(v[1], bool):
                        self._atoms.append(Atom(identifier=v[0], dummy=v[1]))
                    elif isinstance(v[1], (list, tuple)):
                        self._atoms.append(Atom(identifier=v[0], xyz=v[1]))
                    else:
                        raise ValueError("Second argument should be <bool> or <tuple>, not {}".format(type(v[1])))
                elif len(v) == 3:
                    self._atoms.append(Atom(identifier=v[0], xyz=v[1], dummy=v[2]))
                else:
                    raise ValueError("wrong number of Atom arguments: {}, expecting: 1, 2, 3.".format(len(v)))
        else:
            raise TypeError("{0:s} object is not iterable".format(type(values)))

    def unique(self):
        '''Get a list of unique atom specified by unique keyword'''
        return [self.atoms[i] for i in self.unique_labels]

    def nele(self):

        '''Get the total number of electrons in a molecule.'''

        nelectrons = 0
        for atom in self.atoms:
            if atom.atomic_number > 0:
                nelectrons += atom.atomic_number
        return nelectrons - self.charge


    def get_distance(self, atom1, atom2):

        '''Calcualte the distance between two atoms.'''

        dist = 0.0
        for i in range(3):
            dist += (self.atoms[atom1].xyz[i]-self.atoms[atom2].xyz[i])**2
        return sqrt(dist)

    def gamess_rep(self):

        out = ""
        for atom in self.atoms:
            out = out + "{0:<10s} {1:5.1f}\t{2:15.5f}{3:15.5f}{4:15.5f}\n".format(
                atom.symbol, float(atom.atomic_number), atom.xyz[0], atom.xyz[1], atom.xyz[2])
        return out

    def molpro_rep(self):

        out = "geomtyp=xyz\ngeometry={{\n{0:4d}\n{1:<80s}\n".format(len(self.atoms), self.name)
        for atom in self.atoms:
            out = out + "{0:<10s}, {1:15.5f}, {2:15.5f}, {3:15.5f}\n".format(
                atom.symbol, atom.xyz[0], atom.xyz[1], atom.xyz[2])
        out = out + "}\n"
        return out

    def __repr__(self):

        out = "<Molecule(name={}, charge={}, multiplicity={},\n".format(self.name, self.charge, self.multiplicity)
        for atom in self.atoms:
            out += atom.__repr__() + "\n"
        out += ")>"
        return out

    def __str__(self):

        '''Print formatted molecule data.'''

        out = 'Name: {n:<10s} Charge: {c:<10d} Multiplicty: {m:<10d} Electrons: {e:<10d}\n'.format(
              n=self.name, c=self.charge, m=self.multiplicity, e=self.nele())
        out += 'Atoms:\n'
        out += '{0:<10s} {1:^14s}\t{2:^15s}{3:^15s}{4:^15s}\n'.format("Element", "Nuclear Charge", "x", "y", "z")
        for atom in self.atoms:
            out += str(atom)
        return out



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import os
from collections import Counter
from subprocess import Popen, call

from .calculator import Calculator, InputTemplate, parse_objective

from ..basisset import get_l


class Dalton(Calculator):

    'Wrapper for running Dalton program'

    def __init__(self, name='Dalton', **kwargs):

        self.name = name
        super(Dalton, self).__init__(**kwargs)

        self.daltonpath = os.path.dirname(self.executable)

    def parse(self, fname, objective, regularexp=None):
        '''
        Parse a value from the output file ``fname`` based on the
        ``objective``.

        If the value of the ``objective`` is ``regexp`` then the
        ``regularexp`` will be used to parse the file.
        '''

        regexps = {
            'hf total energy': r'^@\s+Final HF energy:\s*(\-?\d+\.\d+)',
            'cisd total energy': r'\d+\s*\d+\s*(\-?\d+\.\d+).*converged',
            'accomplished': r'End of Wave Function Section',
        }

        if objective == 'regexp':
            if regularexp is None:
                raise ValueError("<regularexp> needs to be specified for objective='regexp'")
            toparse = regularexp
        else:
            toparse = regexps.get(objective, None)
            if toparse is None:
                raise ValueError("Specified objective: '{0:s}' not supported".format(objective))

        return parse_objective(fname, toparse)

    def run(self, fname):
        '''
        Run a single job

        Args:
            fname : dict
                A dictionary with keys ``mol`` and ``dal`` and their
                respective file name strings as values

        Returns:
            out : str
                Name of the dalton output file
        '''

        dalbase = os.path.splitext(fname['dal'])[0]
        molbase = os.path.splitext(fname['mol'])[0]

        command = [self.executable] + self.runopts + [dalbase, molbase]
        call(command)

        return dalbase + '_' + molbase + '.out'

    def run_multiple(self, fnames):
        '''
        Spawn two single jobs as paralell processes
        '''

        procs = []
        outputs = []
        for fname in fnames:

            dalbase = os.path.splitext(fname['dal'])[0]
            molbase = os.path.splitext(fname['mol'])[0]
            outputs.append(dalbase + '_' + molbase + '.out')

            command = [self.executable] + self.runopts + [dalbase, molbase]
            process = Popen(command)
            procs.append(process)

        for proc in procs:
            proc.wait()

        return outputs

    def write_input(self, fname, template, basis, mol, core):
        '''
        Write dalton input files: ``fname.dal`` and ``system.mol``

        Args:
            fname : str
                Name of the input file ``.dal``
            template : dict
                Dictionary with templates for the ``dal`` and ``mol``
                with those strings as keys and actual templates as
                values
            basis : dict
                An instance of
                :py:class:`BasisSet <chemtools.basisset.BasisSet>`
                class or a dictionary of
                :py:class:`BasisSet <chemtools.basisset.BasisSet>`
                objects with element symbols as keys
            mol : :py:class:`chemtools.molecule.Molecule`
                Molecule object with the system geometry
            core : str
                Core definition
        '''

        # Dalton uses atomic units for xyz coordinats by default

        daltemplate = template['dal']
        moltemplate = template['mol']

        # loop over different elements (not atoms)
        atomtypes = Counter([a.symbol for a in mol.atoms])
        out = ''
        for symbol, count in atomtypes.items():
            atoms = [a for a in mol.atoms if a.symbol == symbol]
            atombasis = basis[symbol]
            atombasis.sort()
            # get max angular momentum + 1 and construct block string
            maxb = max([get_l(s) for s in atombasis.functions.keys()]) + 1
            block = str(maxb) + ' 1' * maxb
            out += 'Atoms={0:d} Charge={1:.1f} Block={2:s}\n'.format(count,
                                                                    float(atoms[0].atomic_number),
                                                                    block)
            for i, atom in enumerate(atoms, start=1):
                out += '{0:4s} {1:15.8f} {2:15.8f} {3:15.8f}\n'.format(atom.symbol+str(i),
                                                              atom.xyz[0], atom.xyz[1], atom.xyz[2])

            out += atombasis.to_dalton()

        molsubs = {'basis' : out}
        moltemp = InputTemplate(moltemplate)
        dalsubs = {'core' : core}
        daltemp = InputTemplate(daltemplate)

        with open(fname['mol'], 'w') as fmol:
            fmol.write(moltemp.substitute(molsubs))

        with open(fname['dal'], 'w') as fdal:
            fdal.write(daltemp.substitute(dalsubs))

    def __repr__(self):
        return "\n".join(["<Dalton(",
                          "\tname={},".format(self.name),
                          "\tdaltonpath={},".format(self.daltonpath),
                          "\texecutable={},".format(self.executable),
                          "\tscratch={},".format(self.scratch),
                          "\trunopts={},".format(str(self.runopts)),
                          ")>\n"])



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

from __future__ import print_function

import numpy as np
from scipy.special import factorial2
from chemtools.basisset import ncartesian, get_l
from numba import jit, int32, float64
from scipy.linalg import sqrtm, inv


@jit(float64(int32, float64))
def norm(n, a):
    '''
    Calculate the normalization factor for a cartesian gaussian of the form

    .. math::

       G_{n}(x, a, A_{x}) = x^{n}_{A}\exp(-ax^{2}_{A})

    Args:
      n : int
        Power of the preexponential factor
      a : float
        Exponent

    Returns:
      out : float
        Value of the normalization factor
    '''

    # first option
    #return np.sqrt(np.power(2.0*a, n + 0.5, dtype=np.float64)/gamma(n + 0.5), dtype=np.float64)
    # second option
    return np.sqrt(np.power(4.0*a, n)*np.sqrt(2.0*a/np.pi, dtype=np.float64)/factorial2(2*n - 1), dtype=np.float64)
    # third option
    #return np.sqrt(np.power(8.0*a, n)*factorial(n)*np.sqrt(2.0*a/np.pi)/factorial(2*n), dtype=np.float64)

@jit(float64(int32, int32, float64, float64, float64, float64))
def obara_saika(i, j, a, b, Ax, Bx):
    '''
    Calculates the overlap integral between two  cartesian gaussian components using
    the Obara-Saika recurrence relations.

    Args:
      i : int
        Power of the preexponential monomial in the first gaussian
      j : int
        Power of the preexponential monomial in the second gaussian
      a : float
        Exponent of the first gaussian
      b : float
        Exponent of the second gaussian
      Ax : float
        Coordinate of the first gaussian
      Bx : float
        Coordiante of the second gaussian

    Returns:
      out : float
        The value of the :math:`S_{i,j}` integral
    '''

    p = a + b
    mu = a * b / p
    Px = (a * Ax + b * Bx) / p
    X_AB = Ax - Bx
    X_PA = Px - Ax
    X_PB = Px - Bx

    if i == 0 and j == 0:
        return np.sqrt(np.pi / p) * np.exp(-mu * X_AB**2)
    elif i == 0 and j == 1:
        return X_PB * obara_saika(0, 0, a, b, Ax, Bx)
    elif i == 1 and j == 0:
        return X_PA * obara_saika(0, 0, a, b, Ax, Bx)
    elif i == 1 and j == 1:
        return X_PA * obara_saika(0, 1, a, b, Ax, Bx) + \
                obara_saika(0, 0, a, b, Ax, Bx)/(2.0*p)
    elif i == 0 and j > 1:
        return X_PB * obara_saika(0, j - 1, a, b, Ax, Bx) + \
                (j-1)*obara_saika(0, j - 2, a, b, Ax, Bx)/(2.0*p)
    elif i == 1 and j > 1:
        return X_PB * obara_saika(1, j - 1, a, b, Ax, Bx) + \
                (obara_saika(0, j-1, a, b, Ax, Bx) + \
                (j - 1) * obara_saika(1, j-2, a, b, Ax, Bx))/(2.0*p)
    elif i > 1 and j == 0:
        return X_PA * obara_saika(i-1, 0, a, b, Ax, Bx) + \
                (i - 1)*obara_saika(i-2, 0, a, b, Ax, Bx)/(2.0*p)
    elif i > 1 and j == 1:
        return X_PA*obara_saika(i-1, 1, a, b, Ax, Bx) + \
                ((i-1)*obara_saika(i-2, 1, a, b, Ax, Bx) + \
                obara_saika(i-1, 0, a, b, Ax, Bx))/(2.0*p)
    else:
        return X_PA*obara_saika(i-1, j, a, b, Ax, Bx) + \
                ((i-1)*obara_saika(i-2, j, a, b, Ax, Bx) + \
                j*obara_saika(i-1, j-1, a, b, Ax, Bx))/(2.0*p)


def get_basinfo(bases, positions, order='canonical'):
    '''
    Compose a record array with all of the exponents/functions, The record has
    the follwing structure

      * ``exp``,  ``float``, the exponent
      * ``shell``, ``str``, the shell
      * ``l``, ``int``, the angular momentum quantum number
      * ``ix``, ``int``, the power fo the ``x`` monomial
      * ``iy``, ``int``, the power fo the ``y`` monomial
      * ``iz``, ``int``, the power fo the ``z`` monomial
      * ``atom``, ``str``, symbol of the element
      * ``x``, ``float``, the x coordinate
      * ``y``, ``float``, the y coordinate
      * ``z``, ``float``, the z coordinate

    Args:
      bases : list
        A list of ``BasisSet`` objects
      positions : list
        List of 3-tuples with the (x, y, z) coordinates
      order : str
        Order in which the components are sorted, default is ``canonical``, meaning
        [*xx*, *xy*, *xz*, *yy*, *yz*, *zz*] for the *d* shell, other choice is ``gamessus``
        where the highest powers are listed first [*xx*, *yy*, *zz*, *xy*, *xz*, *yz*]

    Returns:
      out : numpy.recarray
        Numpy record array with all available functions
    '''

    dtype = [('exp', np.float64), ('shell', 'S1'), ('l', np.int32),
             ('ix', np.int32), ('iy', np.int32), ('iz', np.int32),
             ('atom', 'S3'), ('x', np.float64), ('y', np.float64), ('z', np.float64)]

    nf = sum([bs.nprimitive(spherical=False) for bs in bases])
    out = np.recarray((nf,), dtype=dtype)

    index = -1
    for basis, atom in zip(bases, positions):
        for shell, fs in basis.functions.items():
            l = get_l(shell)
            for ex in fs['e']:
                for ixyz in xyzlist(l, order=order):
                    index += 1
                    out[index] = (ex, shell, l, ixyz[0], ixyz[1], ixyz[2],
                                  basis.element, atom[0], atom[1], atom[2])
    return out


def xyzlist(l, order='canonical'):
    '''
    Generate an array of :math:`l_x`, :math:`l_y`, :math:`l_z` components of
    cartesian gaussian with a given angular momentum value in canonical order.

    For exampe:
      - ``l = 0`` generates the result ``array([[0, 0, 0]])``
      - ``l = 1`` generates ``array([[1, 0, 0], [0, 1, 0], [0, 0, 1])``
      - etc.

    The functions are coded by triples of powers of *x*, *y*, *z*, namely
    ``[1, 2, 3]`` corresponds to :math:`xy^{2}z^{3}`.

    Args:
      l : int
        Angular momentum value
      order : str
        Order in which the components are sorted, default is ``canonical``, meaning
        [*xx*, *xy*, *xz*, *yy*, *yz*, *zz*] for the *d* shell, other choice is ``gamessus``
        where the highest powers are listed first [*xx*, *yy*, *zz*, *xy*, *xz*, *yz*]

    Returns:
      out ((l+1)*(l+2)/2, 3) : numpy.array
        Array of monomial powers, where columns correspond to *x*, *y*, and *z*
        respectively and rows correspond to functions.
    '''

    ncart = ncartesian(l)
    out = np.zeros((ncart, 3), dtype=np.int32)

    index = 0
    for i in range(l + 1):
        for j in range(i + 1):
            out[index,:] = [l - i, i - j, j]
            index += 1
    if order == 'canonical' or l == 0:
        return out
    elif order == 'gamessus':
        # indices of the second largest value in each row
        imax2 = np.argsort(out, axis=1)[:, 1]
        # indices of largest elements in each row
        imax = np.argmax(out, axis=1)
        # negative sum of two largest values
        nmax2 = -np.sum(np.sort(out, axis=1)[:, 1:], axis=1)
        # negative largest values in each row
        nmax = -np.max(out, axis=1)
        return out[np.lexsort((imax2, imax, nmax2, nmax))]
    else:
        raise ValueError('Unknown <order> value: {}'.format(order))


@jit
def primitive_overlap(exps):
    '''
    Calculate the overlaps between the primitive cartesian gaussians

    Args:
      exps (N,) : numpy.recarray
        Numpy record array with available functions, for details see ``get_exps``

    Returns:
      S (N, N) : numpy.array
        Overlap integral matrix between the primitives
    '''

    # initialize the overlap matrix
    S = np.zeros((exps.size, exps.size), dtype=np.float64)

    # this should be done is some smarter way avoiding the explicit loop indexing
    # maybe np.nditer?

    for i in range(exps.size):

        irow = exps[i]
        iix, iiy, iiz, iexp, xi, yi, zi = irow['ix'], irow['iy'], irow['iz'], irow['exp'], irow['x'], irow['y'], irow['z']

        normi = norm(iix, iexp)*norm(iiy, iexp)*norm(iiz, iexp)

        for j in range(i + 1):
            normj = norm(exps[j]['ix'], exps[j]['exp'])*\
                    norm(exps[j]['iy'], exps[j]['exp'])*\
                    norm(exps[j]['iz'], exps[j]['exp'])

            x = obara_saika(iix, exps[j]['ix'], iexp, exps[j]['exp'], xi, exps[j]['x'])
            y = obara_saika(iiy, exps[j]['iy'], iexp, exps[j]['exp'], yi, exps[j]['y'])
            z = obara_saika(iiz, exps[j]['iz'], iexp, exps[j]['exp'], zi, exps[j]['z'])

            S[i, j] = normi*normj*x*y*z

    S = S + S.T
    np.fill_diagonal(S, S.diagonal()/2.0)
    return S


def primitive_overlap_between(bas1, bas2):

    # initialize the overlap matrix
    S = np.zeros((bas1.size, bas2.size), dtype=np.float64)

    for i in range(bas1.size):
        normi = norm(bas1[i]['ix'], bas1[i]['exp'])*norm(bas1[i]['iy'], bas1[i]['exp'])*norm(bas1[i]['iz'], bas1[i]['exp'])
        for j in range(bas2.size):
            normj = norm(bas2[j]['ix'], bas2[j]['exp'])*norm(bas2[j]['iy'], bas2[j]['exp'])*norm(bas2[j]['iz'], bas2[j]['exp'])

            x = obara_saika(bas1[i]['ix'], bas2[j]['ix'], bas1[i]['exp'], bas2[j]['exp'], bas1[i]['x'], bas2[j]['x'])
            y = obara_saika(bas1[i]['iy'], bas2[j]['iy'], bas1[i]['exp'], bas2[j]['exp'], bas1[i]['y'], bas2[j]['y'])
            z = obara_saika(bas1[i]['iz'], bas2[j]['iz'], bas1[i]['exp'], bas2[j]['exp'], bas1[i]['z'], bas2[j]['z'])

            S[i, j] = normi*normj*x*y*z

    return S


def contraction_matrix(bases):
    '''
    Construct and return the matrix of contraction coefficents

    Args:
      bases : list
        List of BasisSet object

    Returns:
      cc : numpy.array
        Matrix with contraction coefficients
    '''

    # calcualte the number of primitives
    nprim = sum([bs.nprimitive(spherical=False) for bs in bases])
    # calculate the number of contracted functions
    nc = sum([bs.nf(spherical=False) for bs in bases])

    cc = np.zeros((nprim, nc), dtype=np.float64)
    icol = 0

    irow = 0
    for basis in bases:
        for shell, fs in basis.functions.items():
            l = get_l(shell)
            ncart = ncartesian(l)
            nprs = len(fs['e']) * ncart
            for cf in fs['cf']:
                # calculate the indices of rows for all the components
                rowidx = np.asarray([[i * ncart + n  for i in cf['idx']] for n in range(ncart)]).T
                # reshape the contraction coefficients to assign them to elements of the contraction matrix
                cc[rowidx + irow, np.arange(ncart) + icol] = np.repeat(cf['cc'], ncart).reshape(cf.size, ncart)
                icol += ncart
            irow += nprs
    return cc


def completeness_profile(basinfo, cc, zetas):
    '''
    Calculate the completeness profile of each shell of the basis set

    Args:
      zetas : numpy.array
        Scaning exponents

    Returns:
      out : numpy.array
        numpy array with values of the profile (shells, zetas)
    '''

    # calculate the overlap matrix
    Sprim = primitive_overlap(basinfo)
    S = np.dot(cc.T, np.dot(Sprim, cc))
    # calculate the inverse square root of S
    X = inv(sqrtm(S))
    SO = primitive_overlap_between(basinfo, zetas)
    J = np.dot(np.dot(cc, X).T, SO)
    out = np.sum(np.square(J), axis=0)

    return out



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''Module for Complete Basis Set (CBS) Extrapolations.'''

from scipy.optimize import curve_fit
import numpy as np


def extrapolate(x, energy, method, **kwargs):
    '''
    An interface for performing CBS extrapolations using various methods.

    Args:
      x : numpy.array
        A vector of basis set cardinal numbers
      energy : numpy.array
        A vector of corresponding energies
      method : str
        Method/formula to use to perform the extrapolation

    Kwargs:
      Keyword arguments to be passed to the requested extrapolation function
      using the `method` argument
    '''

    methods = {
        'poly': poly,
        'expo': expo,
        'exposqrt': exposqrt,
        'exposum': exposum,
        'uste': uste,
    }

    if len(x) != len(energy):
        raise ValueError("x and energy should have the same size")

    if method in methods.keys():
        eopt, ecov = curve_fit(methods[method](**kwargs), x, energy)
        return eopt
    else:
        raise ValueError("wrong method: {0}, accepted values are: {1}".format(method,
            ", ".join(list(methods.keys()))))


def uste(method="CI"):
    '''
    CBS extrapolation using uniform singlet and triplet pair extrapolation
    (USTE) scheme [Varandas2007]_.

    .. [Varandas2007] Varandas, A. J. C. (2007). "Extrapolating to the one-electron
       basis-set limit in electronic structure calculations. The Journal of
       Chemical Physics, 126(24), 244105. `doi:10.1063/1.2741259 <http://www.dx.doi.org/10.1063/1.2741259>`_

    Args:
      x : int
        Cardinal number of the basis set
      e_cbs : float
        Approximation to the energy value at the CBS limit
      a : float
        Empirical A3 parameter
      method : str
        One of: *ci*, *cc*

    Returns:
      function object
    '''
    def uste_ci(x, e_cbs, a):
        # parameters calibrated for MRCI(Q)
        params = {
            "A05": 0.003769,
            "c": -1.1784771,
            "m": 1.25,
            "alpha": -0.375}

        a5 = params["A05"] + params['c'] * np.power(a, params["m"])
        v = e_cbs + a / np.power(x + params["alpha"], 3) + a5 / np.power(x + params["alpha"], 5)
        return v

    def uste_cc(x, e_cbs, a):
        # parameters calibrated for CC
        params = {
            "A05": 0.1660699,
            "c": -1.4222512,
            "m": 1.0,
            "alpha": -0.375}

        a5 = params["A05"] + params['c'] * np.power(a, params["m"])
        v = e_cbs + a / np.power(x + params["alpha"], 3) + a5 / np.power(x + params["alpha"], 5)
        return v

    if method.lower() == "ci":
        return uste_ci
    elif method.lower() == "cc":
        return uste_cc
    else:
        ValueError("wrong method: {}, accepted values are: 'ci', 'cc'".format(method))


def exposqrt(twopoint=True):
    '''
    Three-point formula for extrapolating the HF reference energy [2]_.

    .. [2] Karton, A., & Martin, J. M. L. (2006). Comment on: “Estimating the
       Hartree-Fock limit from finite basis set calculations” [Jensen F (2005)
       Theor Chem Acc 113:267]. Theoretical Chemistry Accounts, 115, 330–333.
       `doi:10.1007/s00214-005-0028-6 <http://www.dx.doi.org/10.1007/s00214-005-0028-6>`_

    .. math::

       E^{HF}(X) = E_{CBS} + a\cdot \exp(-b\sqrt{X})

    Args:
      twpoint : bool
        A flag marking the use of two point extrapolation with `b=9.0`

    Returns:
      funtion object
    '''

    def exposqrt2(x, e_cbs, a):
        '''
        Two-point formula for extrapolating the HF reference energy, as
        proposed by A. Karton and J. M. L. Martin, Theor. Chem. Acc. 115, 330.
        (2006)

        :math:`E^{HF}(X) = E_{CBS} + A\cdot \exp(-9\sqrt{X})`

        Args:
        x : int
            Cardinal number of the basis set
        e_cbs : float
            Approximation to the energy value at the CBS limit
        a : float
            Pre-exponential empirical parameter
        '''
        return e_cbs + a * np.exp(-9.0 * np.sqrt(x))

    def exposqrt3(x, e_cbs, a, b):
        '''
        Three-point formula for extrapolating the HF reference energy, as
        proposed by A. Karton and J. M. L. Martin, Theor. Chem. Acc. 115, 330.
        (2006)

        :math:`E^{HF}(X) = E_{CBS} + A\cdot \exp(-B\sqrt{X})`

        Args:
        x : int
            Cardinal number of the basis set
        e_cbs : float
            Approximation to the energy value at the CBS limit
        a : float
            Pre-exponential empirical parameter
        b : float
            Exponential empirical parameter
        '''
        return e_cbs + a * np.exp(-b * np.sqrt(x))

    if twopoint:
        return exposqrt2
    else:
        return exposqrt3


def expo():
    '''
    CBS extrapolation formula by exponential Dunning-Feller relation.

    .. math::

       E^{HF}(X) = E_{CBS} + a\cdot\exp(-bX)

    Returns:
      function object
    '''

    def exponential(x, e_cbs, a, b):
        '''
        CBS extrapolation formula by exponential Dunning-Feller relation.

        Args:
        x : int
            Cardinal number of the basis set
        e_cbs : float
            Approximation to the energy value at the CBS limit
        a : float
            Pre-exponential empirical parameter
        b : float
            Exponential empirical parameter
        '''

        return e_cbs + b * np.exp(-a * x)

    return exponential


def exposum():
    '''
    Three point extrapolation through sum of exponentials expression

    .. math::

       E(X) = E_{CBS} + a \cdot\exp(-(X-1)) + b\cdot\exp(-(X-1)^2)

    '''

    def exponentialssum(x, e_cbs, a, b):
        '''
        Three point extrapolation through sum of exponentials expression

        :math:`E(X) = E_{CBS} + A \cdot\exp(-(X-1)) + B\cdot\exp(-(X-1)^2)`

        Args:
        x : int
            Cardinal number of the basis set
        e_cbs : float
            Approximation to the energy value at the CBS limit
        a : float
            Empirical coefficent
        b : float
            Empirical coefficent
        '''
        return e_cbs + a * np.exp(-(x - 1)) + b * np.exp(-(x - 1)**2)

    return exponentialssum


def poly(p=0.0, z=3.0, twopoint=True):
    '''
    CBS extrapolation by polynomial relation.

    .. math::

       E(X) = E_{CBS} + \sum_{i}a_{i}\cdot (X + P)^{-b_{i}}

    Kwargs:
      twpoint : bool
        A flag for choosing the two point extrapolation
      z : float or list of floats
        Order of the polynomial, *default=3.0*
      p : float
        A parameter modifying the cardinal number, *default=0.0*
    '''

    def poly3(x, e_cbs, a, z):
        '''
        Args:
          x : int
            Cardinal number of the basis set
          e_cbs : float
            Approximation to the energy value at the CBS limit
          a : float or list of floats
            Polynomial coefficient
        '''

        a = np.array(a)
        z = np.array(z)
        return e_cbs + np.dot(a, np.power((x + p), -z))

    def poly2(x, e_cbs, a):
        '''
        Args:
          x : int
            Cardinal number of the basis set
          e_cbs : float
            Approximation to the energy value at the CBS limit
          a : float or list of floats
            Polynomial coefficient
        '''

        a = np.array(a)
        zeta = np.array(z)
        return e_cbs + np.dot(a, np.power((x + p), -zeta))

    if twopoint:
        return poly2
    else:
        return poly3



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

from __future__ import print_function
from subprocess import Popen, call
import os

from .calculator import Calculator, InputTemplate, parse_objective


class Psi4(Calculator):
    '''
    Wrapper for the Psi4 program.
    '''

    def __init__(self, name="Psi4", **kwargs):
        self.name = name
        self.inpext = '.dat'
        super(Psi4, self).__init__(**kwargs)

        self.psi4path = os.path.dirname(self.executable)

    def run(self, inpfile):
        '''
        Run a single Psi4 job interactively - without submitting to the queue.
        '''

        outfile = os.path.splitext(inpfile)[0] + ".out"

        command = [self.executable, inpfile] + self.runopts
        call(command)

        return outfile

    def run_multiple(self, inputs):
        '''
        Run a single Psi4 job interactively - without submitting to the queue.
        '''

        procs = []
        outputs = [os.path.splitext(inp)[0] + ".out" for inp in inputs]
        for inpfile, outfile in zip(inputs, outputs):
            opts = []
            opts.extend([self.executable, inpfile] + self.runopts)
            out = open(outfile, 'w')
            process = Popen(opts, stdout=out, stderr=out)
            out.close()
            procs.append(process)

        for proc in procs:
            proc.wait()

        return outputs

    def write_input(self, fname, template, mol=None, basis=None, core=None):
        '''
        Write the Psi4 input to "fname" file based on the information from the
        keyword arguments.

        Args:
            mol : :py:class:`chemtools.molecule.Molecule`
                Molecule object instance
            basis : dict or :py:class:`BasisSet <chemtools.basisset.BasisSet>`
                An instance of :py:class:`BasisSet <chemtools.basisset.BasisSet>` class or a
                dictionary of :py:class:`BasisSet <chemtools.basisset.BasisSet>` objects with
                element symbols as keys
            core : list of ints
                Psi4 core specification
            template : :py:class:`str`
                Template of the input file
            fname : :py:class:`str`
                Name of the input file to be used
        '''

        temp = InputTemplate(template)

        if isinstance(basis, dict):
            bs_str = "".join(x.to_gaussian() for x in basis.values())
        else:
            bs_str = basis.to_gaussian()

        if core is not None:
            core = "core,{0:s}\n".format(",".join([str(x) for x in core]))
        else:
            core = ''

        subs = {
            'geometry' : mol.molpro_rep(),
            'basisset' : bs_str + "\n",
            'core' : core,
        }

        with open(fname, 'w') as inp:
            inp.write(temp.substitute(subs))

    def parse(self, fname, objective, regularexp=None):
        '''
        Parse a value from the output file ``fname`` based on the ``objective``.

        If the value of the ``objective`` is ``regexp`` then the ``regularexp`` will
        be used to parse the file.
        '''

        regexps = {
            'hf total energy'   : r'@RHF Final Energy:\s+(\-?\d+\.\d+)',
            'cisd total energy' : r'\s+\* CISD total energy:\s*(\-?\d+\.\d+)',
            'accomplished'    : r'\*\*\* PSI4 exiting successfully.',
        }

        if objective == 'regexp':
            toparse = regularexp
        else:
            toparse = regexps.get(objective, None)
            if toparse is None:
                raise ValueError("Specified objective: '{0:s}' not supported".format(objective))

        return parse_objective(fname, toparse)

    def __repr__(self):
        return "\n".join(["<Psi4(",
                          "\tname={},".format(self.name),
                          "\tpsi4path={},".format(self.psi4path),
                          "\texecutable={},".format(self.executable),
                          "\tscratch={},".format(self.scratch),
                          "\trunopts={},".format(str(self.runopts)),
                          ")>\n"])



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''
An abstract class that should be subclassed when addding a new code interface.

All the methods should be implemented for the basisopt module to work with the
new code.
'''

import os
import re

from string import Template

from abc import ABCMeta, abstractmethod


class Calculator():
    '''
    Abstract class that should be subclassed when adding a new code interace.
    '''
    __metaclass__ = ABCMeta

    def __init__(self, exevar=None, executable=None, runopts=None,
                 scratch=None):

        self.exevar = exevar

        if executable is None and exevar is None:
            self.exe_found = False
        elif executable is None and os.getenv(self.exevar) is not None:
            self.executable = os.getenv(self.exevar)
            self.exe_found = True
        elif executable is not None and exevar is None:
            self.executable = executable
            self.exe_found = True
        elif executable is not None and exevar is not None:
            exefromvar = os.getenv(self.exevar)
            if os.path.abspath(exefromvar) == os.path.abspath(executable):
                self.executable = exefromvar
                self.exe_found = True
            else:
                raise ValueError('conflicting <executable> and <exevar> options specified: '\
                                 '{0:s} != {1:s}'.format(executable, exefromvar))

        self.runopts = runopts
        self.scratch = scratch

    @property
    def executable(self):
        'Set the executable'
        return self._executable

    @executable.setter
    def executable(self, value):
        if os.path.exists(value):
            if os.path.isdir(value):
                raise OSError('{0} is a directory, not executable'.format(value))
            elif os.path.isfile(value):
                if os.access(value, os.X_OK):
                    self._executable = value
                else:
                    raise OSError('file: {0} is not executable'.format(value))
            else:
                raise OSError('unknown type of the executable: {0}'.format(value))
        else:
            raise ValueError("executable: '{0}' does not exists".format(value))

    @property
    def scratch(self):
        'Return scratch'
        return self._scratch

    @scratch.setter
    def scratch(self, path):

        if path is None:
            self._scratch = os.path.join(os.getenv('HOME'), 'scratch')
        elif os.path.exists(path):
            self._scratch = path
        else:
            raise ValueError("Scratch directory doesn't exist: {}".format(path))

    @abstractmethod
    def parse(self, fname, objective, regexp=None):
        '''
        Parse the value of the ``objective`` from the file ``fname``
        '''
        pass

    @abstractmethod
    def run(self, *args):
        '''
        run a single job
        '''

        pass

    @abstractmethod
    def run_multiple(self, *args):
        '''
        Run multiple jobs
        '''
        pass

    def accomplished(self, fname):
        '''
        Return True if the job completed without errors
        '''

        return self.parse(fname, 'accomplished') is True

    @abstractmethod
    def write_input(self):
        '''
        write the input file in the format of the code used
        '''
        pass


def parse_objective(fname, regexp, reflags=re.I):
    '''
    Wrapper for the parse methods

    Args:
        fname : :py:class:`str`
            Name of the file to be parsed
        regexp : :py:class:`str`
            Regular expression as raw string
        reflags :
            Flags for the regular expression compilation
    '''

    if not os.path.exists(fname):
        raise ValueError("Output file: {0:s} doesn't exist in {1:s}".format(fname, os.getcwd()))

    rec = re.compile(regexp, flags=reflags)
    with open(fname, 'r') as fobj:
        for line in fobj:
            match = rec.search(line)
            if match:
                # check if a group was captured or not
                if match.lastindex is not None and match.lastindex > 0:
                    return float(match.group(1))
                else:
                    return True
        else:
            return None


class InputTemplate(Template):

    'Modified string.Template class to be used for input rendering'

    delimiter = '%'
    idpattern = r'[a-z][_a-z0-9]*'

    def get_keys(self):
        '''
        Parse the string and return a dict with possible keys to substitute.
        For most use case only the `named` fields are interesting.
        '''

        keys = {}
        match = self.pattern.findall(self.template)
        for key, val in self.pattern.groupindex.items():
            keys[key] = [x[val - 1] for x in match if x[val - 1] != '']
        return keys



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

from chemtools.calculators.calculator import Calculator
from chemtools.calculators.gamessus import GamessLogParser
from subprocess import Popen

import os


class Dmft(Calculator):

    '''dmft class'''

    def __init__(self, logfile):

        '''
        Initialize the dmft class object, logfile is the gamess log file name.
        '''

        super(Dmft, self).__init__(**kwargs)
        self.logfile = logfile

    def write_input(self, functional=None, a1=None, b1=None, a2=None, b2=None, printlevel=None, analyze=None):

        '''Write dmft input based on information in the gamess log file.'''

        inputdata = self.parse_gamess()

        if functional:
            inputdata["functional"]  = functional
        if a1:
            inputdata["a1"]          = a1
        if b1:
            inputdata["b1"]          = b1
        if a2:
            inputdata["a2"]          = a2
        if b2:
            inputdata["b2"]          = b2
        if printlevel:
            inputdata["print_level"] = printlevel
        if analyze:
            inputdata["analyze"]     = ".true."

        filebase        = "dmft_fun{0:<s}".format(str(inputdata["functional"]))
        self.inputfile  = filebase + ".inp"
        self.outputfile = filebase + ".out"
        self.nosfile    = filebase + ".nos"

        inp = open(self.inputfile, 'w')
        inp.write("&input\n")
        for key in sorted(inputdata.keys()):
            inp.write("\t{:s}={:s}\n".format(key, str(inputdata[key])))
        inp.write("/\n")
        inp.close()

    def parse_gamess(self):
        '''
        Parse gamess-us log file to get the neccessary data to write dmft input
        and set defaults.
        '''

        gp = GamessLogParser(self.logfile)

        inputdata = {
            "a1"             : 0.0,
            "b1"             : 0.0,
            "a2"             : 0.0,
            "b2"             : 0.0,
            "analyze"        : ".false.",
            "dictfile"       : "'{}'".format(gp.dictionary),
            "exportNOs"      : ".false.",
            "functional"     : 2,
            "loadNOs"        : 1,
            "loadIntegrals"  : 0,
            "nbasis"         : gp.get_number_of_mos(),
            "nuclear_charge" : gp.get_electrons()+gp.get_charge(),
            "print_level"    : 1,
            "restart"        : 0,
            "title"          : "'{}'".format(os.path.splitext(self.logfile)[0]),
            "total_charge"   : gp.get_charge(),
            "twointfile"     : "'{}'".format(gp.twoemofile),
        }
        return inputdata

    def run(self):

        '''Run a single dmft job.'''

        out = open(self.outputfile, 'w')
        process = Popen([self.execpath, self.inputfile], stdout=out, stderr=out)
        process.wait()
        out.close()



# -*- coding: utf-8 -*-

import os
import math
import numpy as np
from sqlalchemy import Column, Integer, String, Float, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property

Base = declarative_base()

dtxyz = np.dtype([('x', 'f8'), ('y', 'f8'), ('z', 'f8')])

class Atom(Base):
    __tablename__ = 'atoms'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    abspath = Column(String)
    output_name = Column(String)
    basisset = Column(String)
    hf_energy = Column(Float)
    ci_energy = Column(Float)
    cc_energy = Column(Float)
    status = Column(String)

    @hybrid_property
    def output(self):
        '''
        Return the absolute path of the output.
        '''
        return os.path.join(self.abspath, self.output_name)

    @hybrid_property
    def input_name(self):
        '''
        Compose the input name from the atom name and the basis set.
        '''
        return "{m:s}_{b:s}.inp".format(m=self.name, b=self.basisset)

    def get_xyz(self, x0=0.0, y0=0.0, z0=0.0):
        '''
        Return atom (x,y,z) coordinates.
        '''

        return np.asarray((x0, y0, z0), dtype=dtxyz)

    def __repr__(self):
        return "<Atom(name={n}, basis={b})>".format(n=self.name, b=self.basisset)

class Dimer(Base):
    __tablename__ = 'dimers'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    abspath = Column(String)
    output_name = Column(String)
    basisset = Column(String)
    hf_energy = Column(Float)
    ci_energy = Column(Float)
    cc_energy = Column(Float)
    status = Column(String)
    r_atom1_atom2 = Column(Float)

    @hybrid_property
    def output(self):
        '''
        Return the absolute path of the output.
        '''
        return os.path.join(self.abspath, self.output_name)

    @hybrid_property
    def input_name(self):
        '''
        Compose the input name from the molecule name, basis set and distance.
        '''
        return "{m:s}_{b:s}_r{r:.2f}.inp".format(m=self.name,
                                                 b=self.basisset,
                                                 r=self.r_atom1_atom2)

    def get_xyz(self, x0=0.0, y0=0.0, z0=0.0):
        '''
        Calculate (x,y,z) coordiantes from internal (r_atom1_atom2).
        '''

        return np.asarray([(x0, y0, z0 - self.r_atom1_atom2/2.0),
                           (x0, y0, z0 + self.r_atom1_atom2/2.0)], dtype=dtxyz)

    def __repr__(self):
        return "<Dimer(name={m}, basis={b}, r_atom1_atom2={raa:5.2f)}>".format(
                m=self.name, b=self.basisset, raa=self.r_atom1_atom2)

class Trimer(Base):
    __tablename__ = 'trimers'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    abspath = Column(String, nullable=True)
    output_name = Column(String)
    basisset = Column(String)
    hf_energy = Column(Float, nullable=True)
    ci_energy = Column(Float, nullable=True)
    cc_energy = Column(Float, nullable=True)
    status = Column(String)
    r_atom_mol = Column(Float)
    gamma = Column(Float)
    r_atom1_atom2 = Column(Float)

    @hybrid_property
    def output(self):
        '''
        Return the absolute path of the output.
        '''
        return os.path.join(self.abspath, self.output_name)

    @hybrid_property
    def input_name(self):
        '''
        Compose the input name from the molecule name, basis set and distance.
        '''
        return "{m:s}_{b:s}_R{R:.2f}_r{r:.2f}_g{g:.2f}.inp".format(m=self.name,
                b=self.basisset, R=self.r_atom_mol, r=self.r_atom1_atom2,
                g=self.gamma)

    def get_xyz(self, x0=0.0, y0=0.0, z0=0.0):
        '''
        Calculate (x,y,z) coordiantes from internal (R_heh2, r_hh, gamma).
        '''

        atom1 = np.asarray((x0, y0, z0), dtype=dtxyz)
        cm = np.asarray((atom1['x'], atom1['y'], atom1['z'] + self.r_atom_mol), dtype=dtxyz)
        atom2 = np.asarray((cm['x'] + 0.5*self.r_atom1_atom2*math.sin(math.radians(self.gamma)),
                    cm['y'],
                    cm['z'] + 0.5*self.r_atom1_atom2*math.cos(math.radians(self.gamma))), dtype=dtxyz)
        atom3 = np.asarray((cm['x'] - 0.5*self.r_atom1_atom2*math.sin(math.radians(self.gamma)),
                    cm['y'],
                    cm['z'] - 0.5*self.r_atom1_atom2*math.cos(math.radians(self.gamma))), dtype=dtxyz)

        if self.gamma not in [0.0, 90.0]:
            for atom in [atom1, atom2, atom3]:
                atom['x'], atom['y'], atom['z'] = atom['z'], atom['x'], atom['y']

        return np.concatenate([a.reshape(1, 3) for a in [atom1, atom2, atom3]], axis=0)

    def __repr__(self):
        return "<Trimer(name={m}, basis={b}, r_atom_mol={r:5.2f}, gamma={g:5.2f}, r_atom1_atom2={raa:5.2f})>".format(
                m=self.name, b=self.basisset, r=self.r_atom_mol, g=self.gamma, raa=self.r_atom1_atom2)

class Tetramer(Base):
    __tablename__ = 'tetramers'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    abspath = Column(String)
    input_name = Column(String)
    output_name = Column(String)
    basisset = Column(String)
    hf_energy = Column(Float)
    ci_energy = Column(Float)
    cc_energy = Column(Float)
    status = Column(String)
    r_mol1_mol2 = Column(Float)
    gamma = Column(Float)
    phi_1 = Column(Float)
    phi_2 = Column(Float)
    r_mol1 = Column(Float)
    r_mol2 = Column(Float)

    @hybrid_property
    def output(self):
        '''
        Return the absolute path of the output.
        '''
        return os.path.join(self.abspath, self.output_name)

    @hybrid_property
    def input_name(self):
        '''
        Compose the input name from the molecule name, basis set and distance.
        '''
        return "{m:s}_{b:s}_R{R:.2f}_r{r2:.2f}_r{r2:.2f}_p{p1:.2f}_p{p2:.2f}_g{g:.2f}.inp".format(m=self.name,
                b=self.basisset, R=self.r_mol1_mol2, r1=self.r_mol1,
                r2=self.r_mol2, p1=self.phi_1, p2=self.phi_2, g=self.gamma)

    def get_xyz(self, x0=0.0, y0=0.0, z0=0.0):
        '''
        Calculate (x,y,z) coordiantes from internal (r_mol1_mol2, r_mol1, r_mol2, phi_1, phi_2, gamma).
        '''

        o = np.asarray((x0, y0, z0))
        cm1 = np.add(o, np.asarray((0.0, 0.0, - self.r_mol1_mol2/2.0)))
        cm2 = np.add(o, np.asarray((0.0, 0.0, + self.r_mol1_mol2/2.0)))

        atom1 = np.asarray((cm1[0] + 0.5*self.r_mol1*math.sin(math.radians(self.phi_1)),
                    cm1[1],
                    cm1[2] + 0.5*self.r_mol1*math.cos(math.radians(self.phi_1))))
        atom2 = np.asarray((cm1[0] - 0.5*self.r_mol1*math.sin(math.radians(self.phi_1)),
                    cm1[1],
                    cm1[2] - 0.5*self.r_mol1*math.cos(math.radians(self.phi_1))))
        atom3 = np.asarray((cm2[0] + 0.5*self.r_mol2*math.sin(math.radians(self.phi_2)),
                    cm2[1],
                    cm2[2] + 0.5*self.r_mol2*math.cos(math.radians(self.phi_2))))
        atom4 = np.asarray((cm2[0] - 0.5*self.r_mol2*math.sin(math.radians(self.phi_2)),
                    cm2[1],
                    cm2[2] - 0.5*self.r_mol2*math.cos(math.radians(self.phi_2))))

        # get the rotation matrix around the z axis
        Rmatrix = rotation_matrix(math.radians(self.gamma), [0, 0, 1])

        atom3 = np.dot(Rmatrix, atom3)
        atom4 = np.dot(Rmatrix, atom4)

        return np.concatenate([a.reshape(1, 3) for a in [atom1, atom2, atom3, atom4]], axis=0)

    def __repr__(self):
        return "<Tetramer(name={n:15s}, basis={b:15s}, r_mol1_mol2={r:5.2f}, phi_1={p1:5.2f}, phi_2={p2:5.2f}>, gamma={g:5.2f})".format(
               n=self.name, b=self.basisset, r=self.r_mol1_mol2, g=self.gamma, p1=self.phi_1, p2=self.phi_2)

def unit_vector(data, axis=None, out=None):
    """Return ndarray normalized by length, i.e. Euclidean norm, along axis.

    >>> v0 = numpy.random.random(3)
    >>> v1 = unit_vector(v0)
    >>> numpy.allclose(v1, v0 / numpy.linalg.norm(v0))
    True
    >>> v0 = numpy.random.rand(5, 4, 3)
    >>> v1 = unit_vector(v0, axis=-1)
    >>> v2 = v0 / numpy.expand_dims(numpy.sqrt(numpy.sum(v0*v0, axis=2)), 2)
    >>> numpy.allclose(v1, v2)
    True
    >>> v1 = unit_vector(v0, axis=1)
    >>> v2 = v0 / numpy.expand_dims(numpy.sqrt(numpy.sum(v0*v0, axis=1)), 1)
    >>> numpy.allclose(v1, v2)
    True
    >>> v1 = numpy.empty((5, 4, 3))
    >>> unit_vector(v0, axis=1, out=v1)
    >>> numpy.allclose(v1, v2)
    True
    >>> list(unit_vector([]))
    []
    >>> list(unit_vector([1]))
    [1.0]

    """
    if out is None:
        data = np.array(data, dtype=np.float64, copy=True)
        if data.ndim == 1:
            data /= math.sqrt(np.dot(data, data))
            return data
    else:
        if out is not data:
            out[:] = np.array(data, copy=False)
        data = out
    length = np.atleast_1d(np.sum(data*data, axis))
    np.sqrt(length, length)
    if axis is not None:
        length = np.expand_dims(length, axis)
    data /= length
    if out is None:
        return data

def rotation_matrix(angle, direction, point=None):
    """Return matrix to rotate about axis defined by point and direction.

    >>> R = rotation_matrix(math.pi/2, [0, 0, 1], [1, 0, 0])
    >>> numpy.allclose(numpy.dot(R, [0, 0, 0, 1]), [1, -1, 0, 1])
    True
    >>> angle = (random.random() - 0.5) * (2*math.pi)
    >>> direc = numpy.random.random(3) - 0.5
    >>> point = numpy.random.random(3) - 0.5
    >>> R0 = rotation_matrix(angle, direc, point)
    >>> R1 = rotation_matrix(angle-2*math.pi, direc, point)
    >>> is_same_transform(R0, R1)
    True
    >>> R0 = rotation_matrix(angle, direc, point)
    >>> R1 = rotation_matrix(-angle, -direc, point)
    >>> is_same_transform(R0, R1)
    True
    >>> I = numpy.identity(4, numpy.float64)
    >>> numpy.allclose(I, rotation_matrix(math.pi*2, direc))
    True
    >>> numpy.allclose(2, numpy.trace(rotation_matrix(math.pi/2,
    ...                                               direc, point)))
    True

    """
    sina = math.sin(angle)
    cosa = math.cos(angle)
    direction = unit_vector(direction[:3])
    # rotation matrix around unit vector
    R = np.diag([cosa, cosa, cosa])
    R += np.outer(direction, direction) * (1.0 - cosa)
    direction *= sina
    R += np.array([[ 0.0,         -direction[2],  direction[1]],
                      [ direction[2], 0.0,          -direction[0]],
                      [-direction[1], direction[0],  0.0]])
    return R




# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''
    GamessReader    : reading gamess binary files.
'''

import os
import sys
import numpy as np
from collections import namedtuple

from .gamessus import GamessLogParser

# add an option to choose which reader should be used for dictionary file
# and sequential files, the options are:
# - wrapped fortran code that need to be compiled and installed or
# - native reader written in python using DictionaryFile and SequentialFile
# classes

# fortran modules nedded for GamessReader Class
# from gamessusfortranext import twoe


class GamessFortranReader(object):
    '''
    Class for holding method for reading gamess binary files:
        $JOB.F08 : two electron integrals over AO's,
        $JOB.F09 : two electron integrals over MO's,
        $JOB.F10 : the dictionary file with 1-e integrals, orbitals etc.,
        $JOB.F15 : GUGA and ORMAS two-electron reduced density matrix,
    '''

    def __init__(self, log):
        self.logfile = log
        i = self.logfile.index("log")
        self.filebase = self.logfile[: i - 1]
        self.twoeaofile = self.filebase + ".F08"
        self.twoemofile = self.filebase + ".F09"
        self.rdm2file = self.filebase + ".F15"
        self.gp = GamessLogParser(log=self.logfile)

    def get_onee_size(self, aos=True):
        '''
        Get the size of the vector holding upper (or lower) triangle
        of a square matrix of size naos or nmos.
        '''
        if aos:
            n = self.gp.get_number_of_aos()
        else:
            n = self.gp.get_number_of_mos()
        return n * (n + 1) // 2

    def get_twoe_size(self, aos=False):
        '''
        Get the size of the 1d vector holding upper (or lower) triangle
        of a supermatrix of size nmos (2RDM and two-electrons integrals).
        '''
        n = self.get_onee_size(aos)
        return n * (n + 1) // 2

    def read_rdm2(self, filename=None, nmo=None):

        '''Read the 2rdm from the gamess-us file'''

        # initialize numpy array to zeros
        rdm2 = np.zeros(self.get_twoe_size(), dtype=float)

        # use gamess module to read the integrals from the file -filename-
        if filename:
            if os.path.exists(filename):
                print("Reading {}".format(filename))
                twoe.integrals.readinao(rdm2, filename)
                return rdm2
            else:
                sys.exit("File '{0:s}' doesn't exist, exiting...".format(filename))
        elif os.path.exists(self.rdm2file):
            twoe.integrals.readinao(rdm2, self.rdm2file)
            return rdm2
        else:
            sys.exit("File '{0:s}' doesn't exist, exiting...".format(self.rdm2file))

    def read_twoeao(self, filename=None, nmo=None):

        '''Read the two electron integrals from the gamess-us file'''

        # initialize numpy array to zeros
        ints = np.zeros(self.get_twoe_size(aos=True), dtype=float)
        # use gamess module to read the integrals from the file -filename-
        if filename:
            if os.path.exists(filename):
                twoe.integrals.readinao(ints, filename)
                return ints
            else:
                sys.exit("File '{0:s}' doesn't exist, exiting...".format(filename))
        elif os.path.exists(self.twoeaofile):
            twoe.integrals.readinao(ints, self.twoeaofile)
            return ints
        else:
            raise OSError("File '{0:s}' doesn't exist, exiting...".format(self.twoeaofile))

    def read_twoemo(self, filename=None, nmo=None):

        '''Read the two electron integrals from the gamess-us file'''

        # initialize numpy array to zeros
        ints = np.zeros(self.get_twoe_size(), dtype=float)
        # use gamess module to read the integrals from the file -filename-
        if filename:
            if os.path.exists(filename):
                twoe.integrals.readinmo(ints, filename)
                return ints
            else:
                sys.exit("File '{0:s}' doesn't exist, exiting...".format(filename))
        elif os.path.exists(self.twoemofile):
            twoe.integrals.readinmo(ints, self.twoemofile)
            return ints
        else:
            sys.exit("File '{0:s}' doesn't exist, exiting...".format(self.twoemofile))


def ijkl(i, j, k, l):
    '''
    Based on the four orbital indices i, j, k, l return the address
    in the 1d vector.
    '''

    ij = max(i, j) * (max(i, j) + 1) / 2 + min(i, j)
    kl = max(k, l) * (max(k, l) + 1) / 2 + min(k, l)

    return max(ij, kl) * (max(ij, kl) + 1) / 2 + min(ij, kl)


def factor(i, j, k, l):
    '''
    Based on the orbitals indices return the factor that takes into account
    the index permutational symmetry.
    '''

    if i == j and k == l and i == k:
        fijkl = 1.0
    elif i == j and k == l:
        fijkl = 2.0
    elif (i == k and j == l) or (i == j and i == k) or (j == k and j == l) or (i == j or k == l):
        fijkl = 4.0
    else:
        fijkl = 8.0
    return fijkl


def print_twoe(twoe, nbf):
    '''Print the two-electron integrals.'''

    ij = 0
    for i in xrange(nbf):
        for j in xrange(i + 1):
            ij += 1
            kl = 0
            for k in xrange(nbf):
                for l in xrange(k + 1):
                    kl += 1
                    if ij >= kl:
                        if abs(twoe[ijkl(i, j, k, l)]) > 1.0e-10:
                            print("{0:3d}{1:3d}{2:3d}{3:3d} {4:25.14f}".format(
                                i, j, k, l, twoe[ijkl(i, j, k, l)]))


#"""
#BinaryFile: A class for accessing data to/from large binary files
#=================================================================
#
#The data is meant to be read/write sequentially from/to a binary file.
#One can request to read a piece of data with a specific type and shape
#from it.  Also, it supports the notion of Fortran and C ordered data,
#so that the returned data is always well-behaved (C-contiguous and
#aligned).
#
#This is also seeking capable.
#
#:Author:   Francesc Alted
#:Contact:  faltet@pytables.org
#:Created:  2010-03-18
#:Acknowledgment: Funding for the development of this code is provided
#through the Norwegian Research Council VAUUAV project #184724, 2010
#
#"""

class BinaryFile(object):

    """Class representing a binary file (C or Fortran ordered)."""

    def __init__(self, filename, mode="r", order="fortran"):
        """Open the `filename` for write/read binary files.

        The `mode` can be 'r', 'w' or 'a' for reading (default),
        writing or appending.  The file will be created if it doesn't
        exist when opened for writing or appending; it will be
        truncated when opened for writing.  Add a '+' to the mode to
        allow simultaneous reading and writing.

        `order` specifies whether the file is is written in 'fortran'
        or 'c' order.
        """
        self.mode = mode + "b"
        self.file = open(filename, mode=self.mode, buffering=1)
        """The file handler."""
        if order not in ['fortran', 'c']:
            raise ValueError("order should be either 'fortran' or 'c'.")
        self.order = order
        """The order for file ('c' or 'fortran')."""

    def read(self, dtype, shape=(1,)):
        """Read an array of `dtype` and `shape` from current position.

        `shape` must be any tuple made of integers or even () for scalars.

        The current position will be updated to point to the end of
        read data.
        """
        if not hasattr(dtype, "itemsize"):
            dtype = np.dtype(dtype)
        if type(shape) is int:
            shape = (shape,)
        if type(shape) is not tuple:
            raise ValueError("shape must be a tuple")
        length = dtype.itemsize
        if shape is not ():
            length *= np.array(shape).prod()

        # Correct the shape in case dtype is multi-dimensional
        shape = shape + dtype.shape

        if shape in (1, (1,)):
            order = "c"
        else:
            order = self.order

        length = int(length)

        # Read the data from file
        data = self.file.read(length)
        if len(data) < length:
            raise EOFError("Asking for more data than available in file.")
        # Convert string into a regular array
        data = np.ndarray(shape=shape, buffer=data, dtype=dtype.base)

        # If original data file is in fortran order, reverse the shape first
        if order == "fortran":
            shape = [i for i in shape[::-1]]
        data = data.reshape(shape)

        if shape == ():
            # Retrieve the scalar out of the 0-dim array
            data = data[()]
        else:
            # If original data file is in fortran order, do a transpose.
            # As the shape was reversed previously, we get the original
            # shape again.
            if order == "fortran":
                data = data.transpose().copy()
            # Do an additional copy just in case the array is not
            # wel-behaved (i.e., it is not aligned or not contiguous).
            elif not data.flags.behaved:
                data = data.copy()
        return data

    def write(self, arr):
        """Write an `arr` to current position.

        The current position will be updated to point to the end of
        written data.
        """
        # Transpose data if case we need to
        if (self.order == "fortran") != (arr.flags.fortran):
            arr = arr.transpose().copy()
        # Write the data to file
        self.file.write(arr.data)

    def seek(self, offset, whence=0):
        """Move to new file position.

        Argument offset is a byte count.  Optional argument whence
        defaults to 0 (offset from start of file, offset should be >=
        0); other values are 1 (move relative to current position,
        positive or negative), and 2 (move relative to end of file,
        usually negative, although many platforms allow seeking beyond
        the end of a file).  If the file is opened in text mode, only
        offsets returned by tell() are legal.  Use of other offsets
        causes undefined behavior.
        """
        self.file.seek(offset, whence)

    def tell(self):
        "Returns current file position, an integer (may be a long integer)."
        return self.file.tell()


class SequentialFile(BinaryFile):

    def __init__(self, filename, logfile=None):
        '''
        Initialize the class with the superclass method.
        '''
        super(SequentialFile, self).__init__(filename)
        if logfile is None:
            logfile = os.path.splitext(filename)[0] + ".log"
        glp = GamessLogParser(log=logfile)
        self.nao = glp.get_number_of_aos()
        self.nmo = glp.get_number_of_mos()
        self.core = glp.get_number_of_core_mos()
        if self.nao < 255:
            self.large_labels = False
        else:
            self.large_labels = True

    def ijkl(self, i, j, k, l):
        '''
        Based on the four orbital indices i,j,k,l return the address
        in the 1d vector.
        '''

        ij = max(i, j) * (max(i, j) - 1) // 2 + min(i, j)
        kl = max(k, l) * (max(k, l) - 1) // 2 + min(k, l)
        return max(ij, kl) * (max(ij, kl) - 1) // 2 + min(ij, kl) - 1

    def get_index_buffsize(self, buff_size, int_size):
        ''' Return the index buffer size for reading 2-electron integrals'''

        if int_size == 4:
            if self.large_labels:
                return 2 * buff_size
            else:
                return buff_size
        elif int_size == 8:
            if self.large_labels:
                return buff_size
            else:
                return (buff_size + 1) // 2
        else:
            raise ValueError('wrong "int_size": {}'.format(int_size))

    def readseq(self, buff_size=15000, int_size=8, mos=False,
                skip_first=False):
        '''
        Read FORTRAN sequential unformatted file with two-electron
        quantities:

        - two electron integrals over AO's: .F08 file
        - two electron integrals over MO's: .F09 file
        - elements of the two particle density matrix: .F15 file

        Args:
            buff_size : int
                size of the buffer holding values to be read, in
                gamessus it is stored under ``NINTMX`` variable and in
                Linux version is equal to 15000 which is the default
                value
            large_labels : bool
                a flag indicating if large labels should were used, if
                largest label ``i`` (of the MO) is ``i<255`` then
                ``large_labels`` should be ``False``
                (case ``LABSIZ=1`` in gamessus), otherwise set to ``True``
                (case ``LABSIZ=2`` in gamess(us),
            skip_first : bool
                skips the first record of the file is set to True,

        Returns:
            numpy 1D array holding the values
        '''

        indexBuffSize = self.get_index_buffsize(buff_size, int_size)

        self.seek(0)
        if mos:
            nmo = self.nmo - self.core
            nt = self.nmo * (self.nmo + 1) // 2
            if skip_first:
                #n1 = nmo*(nmo+1)/2
                #self.seek(8+8*n1)
                self.seek(self.tell() + 4)
                Hoe = self.read('f8', shape=(nt, ))
                # advance four trailing bytes
                self.seek(self.tell() + 4)

        else:
            nt = self.nao * (self.nao + 1) // 2

        ints = np.zeros(nt * (nt + 1) // 2, dtype=float, order='F')

        int_type = np.dtype('i' + str(int_size))
        index_buffer = np.zeros(indexBuffSize, dtype=int_type, order='F')
        value_buffer = np.zeros(buff_size, dtype=float, order='F')

        length = 1
        while length > 0:
            self.seek(self.tell() + 4)
            length = int(self.read(int_type))
            if length > buff_size:
                raise ValueError('the read record length: {0:10d} is greater that the buffer size {1:10d}'.format(int(length), buff_size))

            index_buffer = self.read(int_type, shape=(indexBuffSize, ))
            value_buffer = self.read('f8', shape=(buff_size, ))

            for m in range(1, abs(length) + 1):
                if int_size == 4:
                    if self.large_labels:
                        label1 = int(index_buffer[2 * m - 1])
                        label2 = int(index_buffer[2 * m])
                        i = label1 >> 16
                        j = label1 & 65535
                        k = label2 >> 16
                        l = label2 & 65535
                    else:
                        label = int(index_buffer[m - 1])
                        i = label >> 24
                        j = label >> 16 & 255
                        k = label >>  8 & 255
                        l = label       & 255
                elif int_size == 8:
                    if self.large_labels:
                        label = int(index_buffer[m - 1])
                        i = label >> 48
                        j = label >> 32 & 65535
                        k = label >> 16 & 65535
                        l = label       & 65535
                    else:
                        if m % 2 == 0:
                            label = int(index_buffer[m // 2 - 1])
                            i = label >> 24 & 255
                            j = label >> 16 & 255
                            k = label >>  8 & 255
                            l = label       & 255
                        else:
                            label = int(index_buffer[m // 2])
                            i = label >> 56 & 255
                            j = label >> 48 & 255
                            k = label >> 40 & 255
                            l = label >> 32 & 255
                ints[self.ijkl(i, j, k, l)] = value_buffer[m - 1]
            self.seek(self.tell() + 4)
        return ints

    def read_ci_coeffs(self):
        '''
        Read CI coefficients from file NFT12 and return them as a numpy array of floats.
        '''

        self.seek(4)
        nstates = self.read('i8')
        nconfs = self.read('i8')

        title = self.read('a80', shape=(2,))

        # advance 4 bytes for the end of the record and additional
        # 4 for the start of the new record
        self.seek(self.tell() + 8)
        return self.read('f8', shape=(nconfs * nstates,))


class GamessReader(object):
    '''
    Class for holding method for reading gamess binary files:
        - $JOB.F08 : two electron integrals over AO's,
        - $JOB.F09 : two electron integrals over MO's,
        - $JOB.F15 : GUGA and ORMAS two-electron reduced density matrix,

    TODO:
        CI coefficients, and CI hamiltonian matrix elements.
    '''
    # add an option to choose which reader should be used for dictionary file
    # and sequential files, the options are:
    # - wrapped fortran code that need to be compiled and installed or
    # - native reader written in python using DictionaryFile and SequentialFile
    # classes

    def __init__(self, log):
        self.logfile = log
        i = self.logfile.index("log")
        self.filebase = self.logfile[:i-1]
        self.datfile = self.filebase + ".dat"
        self.twoeaofile = self.filebase + ".F08"
        self.twoemofile = self.filebase + ".F09"
        self.dictionary = self.filebase + ".F10"
        self.rdm2file = self.filebase + ".F15"
        self.gp = GamessLogParser(log=self.logfile)

    def get_onee_size(self, aos=True):
        '''
        Get the size of the vector holding upper (or lower) triangle
        of a square matrix of size naos or nmos.
        '''
        if aos:
            n = self.gp.get_number_of_aos()
        else:
            n = self.gp.get_number_of_mos()
        return n * (n + 1) // 2

    def get_twoe_size(self):
        '''
        Get the size of the 1d vector holding upper (or lower) triangle
        of a supermatrix of size nmos (2RDM and two-electrons integrals).
        '''
        n = self.get_onee_size(aos=False)
        return n * (n + 1) // 2

    def read_rdm2(self, filename=None, nmo=None):

        '''Read the 2rdm from the gamess-us file'''

        # initialize numpy array to zeros
        rdm2 = np.zeros(self.get_twoe_size(), dtype=float)

    def read_twoeao(self, filename=None, nmo=None):

        '''Read the two electron integrals from the gamess-us file'''

        ints = np.zeros(self.get_twoe_size(), dtype=float)


rec = namedtuple('record', ['name', 'dtype'])
records = {
    1: rec("atomic coordinates", "f8"),
    2: rec("enrgys", "f8"),
    3: rec("gradient vector", "f8"),
    4: rec("hessian matrix", "f8"),
    5: rec("not used", ""),
    6: rec("not used", ""),
    7: rec("ptr", "f8"),
    8: rec("dtr", "f8"),
    9: rec("ftr", "f8"),
   10: rec("gtr", "f8"),
   11: rec("bare nucleus", "f8"),
   12: rec("overlap", "f8"),
   13: rec("kinetic energy", "f8"),
   14: rec("alpha fock matrix", "f8"),
   15: rec("alpha orbitals", "f8"),
   16: rec("alpha density matrix", "f8"),
   17: rec("alpha energies or occupation numbers", "f8"),
   18: rec("beta fock matrix", "f8"),
   19: rec("beta orbitals", "f8"),
   20: rec("beta density matrix", "f8"),
   21: rec("beta energies or occupation numbers", "f8"),
   22: rec("error function interpolation table", "f8"),
   23: rec("old alpha fock matrix", "f8"),
   24: rec("older alpha fock matrix", "f8"),
   25: rec("oldest alpha fock matrix", "f8"),
   26: rec("old beta fock matrix", "f8"),
   27: rec("older beta fock matrix", "f8"),
   28: rec("odest beta fock matrix", "f8"),
   29: rec("vib 0 gradient in FORCE", "f8"),
   30: rec("vib 0 alpha orbitals in FORCE", "f8"),
   31: rec("Vib 0 beta  orbitals in FORCE", "f8"),
   32: rec("Vib 0 alpha density matrix in FORCE", "f8"),
   33: rec("Vib 0 beta  density matrix in FORCE", "f8"),
   34: rec("dipole derivative tensor in FORCE", "f8"),
   35: rec("frozen core Fock operator", "f8"),
   36: rec("RHF/UHF/ROHF Lagrangian", "f8"),
   37: rec("floating point part of common block /OPTGRD/", "f8"),
   38: rec("integer part of common block /OPTGRD/", "i8"),
   39: rec("ZMAT of input internal coords", "f8"),
   40: rec("IZMAT of input internal coords", "i8"),
   41: rec("B matrix of redundant internal coords", "f8"),
   42: rec("pristine core Fock matrix in MO basis (see 87)", "f8"),
   43: rec("Force constant matrix in internal coordinates", "f8"),
   44: rec("SALC transformation", "f8"),
   45: rec("symmetry adapted Q matrix", "f8"),
   46: rec("S matrix for symmetry coordinates", "f8"),
   47: rec("ZMAT for symmetry internal coords", "f8"),
   48: rec("IZMAT for symmetry internal coords", "i8"),
   49: rec("B matrix", "f8"),
   50: rec("B inverse matrix", "f8"),
   69: rec("alpha Lowdin populations", "f8"),
   70: rec("beta Lowdin populations", "f8"),
   71: rec("alpha orbitals during localization", "f8"),
   72: rec("betha orbitals during localization", "f8"),
   73: rec("alpha localization transformation", "f8"),
   74: rec("beta localization transformation", "f8"),
   95: rec("x dipole integrals in AO basis", "f8"),
   96: rec("y dipole integrals in AO basis", "f8"),
   97: rec("z dipole integrals in AO basis", "f8"),
  251: rec("static polarizability tensor alpha", "f8"),
  252: rec("X dipole integrals in MO basis", "f8"),
  253: rec("Y dipole integrals in MO basis", "f8"),
  254: rec("Z dipole integrals in MO basis", "f8"),
  255: rec("alpha MO symmetry labels", "S8"),
  256: rec("beta MO symmetry labels", "S8"),
  286: rec("oriented localized molecular orbitals", "f8"),
  379: rec("Lz integrals", "f8"),
}


class DictionaryFile(BinaryFile):
    '''
    Wrapper for reading GAMESS(US) dictionary file (\*.F10).
    '''

    def __init__(self, filename, irecln=4090, int_size=8):
        """

        Args:
          irecln: int
            Rrecord length that is used by GAMESS(US) when
            writing the dicitonary file. It is a platform dependent
            variable that is called IRECLN in the GAMESS(US) code. In
            GAMESS(US) it is set by the function NRASIZ(UNIT) in iolib.src
            and for dictionary file (unit=10) and UNX it is equal to 4090,
            for all other files and UNX it is equal to 2048. If you use
            GAMESS(US) on a different platform check the NRASIZ(UNIT)
            function for the proper value and supply it when instantiating
            the class.
          int_size: int
            Integer size (in bytes) that the GAMESS(US) was compiled with.
        """
        super(DictionaryFile, self).__init__(filename)

        self.irecln = irecln
        self.int_size = int_size
        # read the first record with the information about the
        # structure of the dictionary file
        self.irecst = self.read(np.dtype('i' + str(self.int_size)))
        self.ioda = self.read(np.dtype('i' + str(self.int_size)), shape=(950,))
        self.ifilen = self.read(np.dtype('i' + str(self.int_size)), shape=(950,))
        self.iss = self.read(np.dtype('i' + str(self.int_size)))
        self.ipk = self.read(np.dtype('i' + str(self.int_size)))

    def read_record(self, nrec, dtype=None):
        '''
        Read a logical record 'rec' from the dictionary file and return a numpy
        array of type defined in the 'records' list, and size defined through
        'self.ifilen' array.
        '''

        if self.ioda[nrec-1] < 0:
            raise IOError("Record {0} was not previously written, IODA[{0}]={1}".format(nrec, self.ioda[nrec-1]))

        self.seek(8 * self.irecln * (int(self.ioda[nrec - 1]) - 1))
        if dtype is not None:
            return self.read(dtype, shape=(self.ifilen[nrec - 1],))
        else:
            return self.read(records[nrec].dtype, shape=(self.ifilen[nrec - 1],))


def tri2full(vector, sym=True):
    '''
    Convert a triagonal matrix whose elements are stored in the `vector` into a
    rectangular matrix of the shape given by `shape` tuple.
    '''

    # get the shape of the symmetric matrix from solving n^2 + n - 2s = 0 equation
    # where n is the number of rows/columns and s is the size of 1D vector

    n = int((np.sqrt(8.0 * vector.size + 1) - 1.0) / 2.0)
    matrix = np.zeros((n, n), dtype=float, order='F')

    ij = -1
    for i in range(n):
        for j in range(i + 1):
            ij += 1
            if sym:
                matrix[i, j] = matrix[j, i] = vector[ij]
            else:
                matrix[i, j] = vector[ij]
                matrix[j, i] = -vector[ij]

    return matrix



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

from __future__ import division, print_function

import json
import re

import numpy as np
from mendeleev import element

CFDTYPE = [('idx', np.int32), ('cc', np.float64)]
ORBITALS = ('s', 'p', 'd', 'f', 'g', 'h', 'i', 'k')


def get_l(shell):
    'Return the orbital angular momentum quantum number for a given subshell'

    if shell in ORBITALS:
        return ORBITALS.index(shell.lower())
    else:
        raise ValueError('"{}" is not a proper shell label'.format(shell))


def parse_basis(string, fmt=None):
    '''
    A wrapper for parsing the basis sets in different formats.

    Args:
      string : str
        A string with the basis set

      fmt : str
        Format in which the basis set is specified

    Returns:
      out : dict
        A dictionary of parsed basis sets with element symbols as keys and
        basis set functions as values
    '''

    if fmt == 'gamessus':
        return parse_gamessus_basis(string)
    elif fmt == 'gaussian':
        return parse_gaussian_basis(string)
    elif fmt == 'molpro':
        return parse_molpro_basis(string)
    else:
        formats = ['molpro', 'gamessus', 'gaussian']

        raise ValueError("Can only parse <fmt>: {}".format(", ".join(formats)))


def parse_molpro_basis(string):
    '''
    Parse basis set from a string in Molpro format.
    '''

    bas_re = re.compile(r'basis\s*=\s*\{(.*?)\}', flags=re.DOTALL | re.I)

    m = bas_re.search(string)
    if m:
        lines = m.group(1).split("\n")
    else:
        raise ValueError('basis string: "basis={*}" not found')

    start = [
        i
        for i, line in enumerate(lines)
        if line.split(",")[0].lower() in ORBITALS
    ]

    if not start:
        return None

    startstop = [(start[i], start[i + 1]) for i in range(len(start) - 1)]
    startstop.append((start[-1], len(lines)))

    bs = {}
    for i in startstop:
        at_symbol, shell = parse_molpro_shell(lines[i[0]],
                                              lines[i[0] + 1: i[1]])
        if at_symbol in bs:
            bs[at_symbol] = dict(list(bs[at_symbol].items()) +
                                 list(shell.items()))
        else:
            bs[at_symbol] = shell
    return bs


def parse_molpro_shell(expsline, coeffs):
    '''
    Parse functions of one shell in molpro format.
    '''

    real = re.compile(r'[dD]')
    # remove empty strings and whitespace line breaks and tabs
    coeffs = [x.strip() for x in coeffs if x.strip() not in ['', '\n', '\t']]

    fs = {}

    shell = expsline.split(",")[0].lower()
    at_symbol = expsline.split(",")[1].strip().capitalize()
    exps = np.array([float(real.sub('E', x))
                     for x in expsline.rstrip(";").split(",")[2:]])

    fs[shell] = {'e': exps, 'cf': []}
    if not coeffs:
        for i in range(len(exps)):
            fs[shell]['cf'].append(np.array([tuple([i, 1.0])], dtype=CFDTYPE))
    else:
        for line in coeffs:
            lsp = line.rstrip(";").split(",")
            if lsp[0] == "c":
                i, j = [int(x) for x in lsp[1].split(".")]
                coeffs = [float(real.sub('E', x)) for x in lsp[2:]]
                fs[shell]['cf'].append(np.array(list(zip(list(range(i - 1, j)),
                                                coeffs)), dtype=CFDTYPE))
    return at_symbol, fs


def parse_ecp(ecpstring):

    ecp_re = re.compile(r'\!\s*Effective core Potentials.*-{25}\s*\n(.*?)\n\s*\n', flags=re.DOTALL)

    lines = ecpstring.split("\n")

    start = [
        i
        for i, line in enumerate(lines)
        if line.split(",")[0].lower() == 'ecp'
    ]

    if not start:
        return None

    startstop = [(start[i], start[i + 1]) for i in range(len(start) - 1)]
    startstop.append((start[-1], len(lines)))

    ecp = {}
    for i in startstop:
        ecp = dict(list(ecp.items()) +
                   list(parse_coeffs(lines[i[0]: i[1]]).items()))
    return ecp


def parse_coeffs(lines):

    firstl = lines[0].replace(';', '').split(',')
    element = firstl[1].strip().capitalize()
    nele = firstl[2]
    lmax = firstl[3]

    liter = iter(x for x in lines[1:] if x != '')

    ecp = {element: {"nele": nele, "lmax": lmax, "shells": []}}

    while True:
        try:
            temp = next(liter)
        except StopIteration:
            break
        nlmax = int(temp.split(";")[0])
        comment = temp.split(";")[1].replace("!", "")
        tt = {'comment': comment, 'parameters': []}
        for _ in range(nlmax):
            param = next(liter).replace(";", "").split(",")
            tt['parameters'].append({'m': float(param[0]),
                                     'gamma': float(param[1]),
                                     'c': float(param[2])})
        ecp[element]['shells'].append(tt)
    return ecp


def parse_gaussian_basis(string):
    '''
    Parse the basis set into a list of dictionaries from a string in
    gaussian format.
    '''

    shellre = re.compile(r'^\s*(?P<shells>[SPDFGHILspdfghil]+)\s*(?P<nf>[1-9]+)\s*(?P<scale>\-?\d+\.\d+)')
    out = {}
    for i, item in enumerate(string.split('****\n')):
        if len(item) > 5:
            atomline, basis = item.split('\n', 1)
            atom = atomline.split()[0]
            bslines = basis.split('\n')
            functions = {}
            for i, line in enumerate(bslines):
                match = shellre.search(line)
                if match:
                    shells, nf, scale = match.group('shells').lower(), match.group('nf'), match.group('scale')
                    exps, indxs, coeffs = parse_gaussian_function(bslines[i + 1: i + int(nf) + 1])
                    for shell, cc in zip(shells, coeffs.T):
                        if shell in functions:
                            sexp, idxs, idxo = merge_exponents(functions[shell]['e'], exps)
                            functions[shell]['e'] = sexp * float(scale)**2
                            for cf in functions[shell]['cf']:
                                cf['idx'] = idxs[cf['idx']]
                            newcf = np.array(list(zip(idxo[indxs], cc)),
                                             dtype=CFDTYPE)
                            functions[shell]['cf'].append(newcf)
                        else:
                            functions[shell] = {'cf': [], 'e': exps}
                            functions[shell]['cf'].append(np.array(list(zip(indxs, cc)), dtype=CFDTYPE))
                out[atom] = functions
    return out


def parse_gaussian_function(lines):
    '''
    Parse a basis set function information from list of strings into
    three lists containg: exponents, indices, coefficients.

    Remeber that python doesn't recognise the `1.0d-3` format where `d` or
    `D` is used to the regex subsitution has to take care of that.
    '''

    real = re.compile(r'[dD]')

    indxs = np.arange(len(lines), dtype=np.int32)
    exps = np.array([float(real.sub('E', line.split()[0])) for line in lines],
                    dtype=np.float64)
    coeffs = np.array([[float(real.sub('E', x))
                        for x in line.split()[1:]] for line in lines],
                      dtype=np.float64)
    return (exps, indxs, coeffs)


def parse_gamessus_basis(string):
    '''
    Parse the basis set into a list of dictionaries from a string in
    gamess format.
    '''

    bas_re = re.compile(r'\$DATA\n(.*?)\$END', flags=re.DOTALL | re.IGNORECASE)

    m = bas_re.search(string)
    if m:
        basisstr = m.group(1)
    else:
        raise ValueError("basis not found, should be inside '$DATA' and '$END' group")

    pat = re.compile(r'^\s*(?P<shells>[SPDFGHILspdfghil]+)\s*(?P<nf>[0-9]+)')
    res = {}

    for item in basisstr.split('\n\n'):
        if len(item) > 0:
            atom, basis = item.split('\n', 1)
            if atom != "":
                functions = {}
                elem = element(atom.capitalize())
                bslines = basis.split("\n")
                for i, line in enumerate(bslines):
                    match = pat.search(line)
                    if match:
                        shells, nf = match.group("shells").lower(), match.group("nf")
                        exps, indxs, coeffs = parse_gamessus_function(bslines[i + 1: i + int(nf) + 1])
                        if shells in ['L', 'l']:
                            shells = ['s', 'p']
                        for shell, cc in zip(shells, coeffs.T):
                            if shell in functions:
                                sexp, idxs, idxo = merge_exponents(functions[shell]['e'], exps)
                                functions[shell]['e'] = sexp
                                for cf in functions[shell]['cf']:
                                    cf['idx'] = idxs[cf['idx']]
                                newcf = np.array(list(zip(idxo[indxs-1], cc)),
                                                 dtype=CFDTYPE)
                                functions[shell]['cf'].append(newcf)
                            else:
                                functions[shell] = {'cf': [], 'e': exps}
                                functions[shell]['cf'].append(np.array(list(zip(indxs - 1, cc)), dtype=CFDTYPE))
                res[elem.symbol] = functions
    return res


def parse_gamessus_function(lines):
    '''
    Parse a basis set function information from list of strings into
    three lists containg: exponents, indices, coefficients.

    Remeber that python doesn't recognise the `1.0d-3` format where `d` or
    `D` is used to the regex subsitution has to take care of that.
    '''

    real = re.compile(r'[dD]')

    indxs = np.array([int(line.split()[0]) for line in lines], dtype=np.int32)
    exps = np.array([float(real.sub('E', line.split()[1])) for line in lines],
                    dtype=np.float64)
    coeffs = np.array([[float(real.sub('E', x)) for x in line.split()[2:]]
                        for line in lines], dtype=np.float64)
    return (exps, indxs, coeffs)


def merge_exponents(a, b):
    '''
    Concatenate the arrays `a` and `b` using only the unique items from both
    arrays

    Args:
      a : numpy.array
      b : numpy.array

    Returns:
      res : 3-tuple of numpy arrays
        - res[0] sorted union of `a` and `b`
        - res[1] indices of `a` items in res[0]
        - res[2] indices of `b` items in res[0]
    '''

    u = np.union1d(a, b)[::-1]
    idxa = np.where(np.in1d(u, a))[0]
    idxb = np.where(np.in1d(u, b))[0]
    return u, idxa, idxb


class NumpyEncoder(json.JSONEncoder):

    def default(self, obj):
        '''
        If input object is an ndarray it will be converted into a dict
        holding the data and dtype.
        '''

        if isinstance(obj, np.ndarray):

            dtype = 'CFDTYPE' if obj.dtype == CFDTYPE else str(obj.dtype)
            return dict(data=obj.tolist(), dtype=dtype)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

"""\
``basisopt`` is a module containing flexible methods for optimizing primitive
exponents of basis sets"""

from __future__ import division, print_function

import atexit
import datetime
import os
import pprint
import random
import string
import sys
import time
from collections import OrderedDict
import numpy as np
from scipy.optimize import minimize

from chemtools.basisparse import ORBITALS
from chemtools.basisset import BasisSet, get_num_params, sliceinto


class BSOptimizer(object):
    '''
    Basis Set Optimizer class is a convenient wrapper for optimizing primitive
    exponents of Gaussian basis sets using different code interfaces for
    performing the actual electronic structure calculations.

    Args:
        objective : str or callable
            Name of the objective using which the optimization will be
            evaluated, if it is a callable/function it should take the output
            name as argument

        code : :py:class:`Calculator <chemtools.calculators.calculator.Calculator>`
            Subclass of the :py:class:`Calculator <chemtools.calculators.calculator.Calculator>`
            specifying the electronic structure program to use

        mol : :py:class:` Molecule <chemtools.molecule.Molecule>`
            Molecule specifying the system

        staticbs : dict or :py:class:`BasisSet <chemtools.basisset.BasisSet>`
            Basis set or ``dict`` of basis set with basis sets whose exponents
            are not going to be optimized

        fsopt : dict
            A dictionary specifying the basis set to be optimized, the keys
            should be element/atom symbol and the values should contain lists
            of 4-tuples composed of: shell, type, number of functions and
            parameters/exponents, for example,

            >>> fs = {'H' : [('s', 'et', 10, (0.5, 2.0))]}

            describes 10 ``s``-type exponents for H generated from even
            tempered formula with parameters 0.5 and 2.0

        optalg : dict
            A dictionary specifying the optimization algorithm and its options

        fname : str
            Name of the job/input file for the single point calculator

        uselogs : bool
            Use natural logarithms of exponents in optimization rather than
            their values, this option is only relevant if functions asre given
            as ``exp`` or ``exponents``

        regexp : str
            Regular expression to use in search for the objective if
            ``objective`` is regexp

        runcore : bool
            Flag to mark wheather to run separate single point jobs with
            different numbers of frozen core orbitals to calculate core energy

        penalize : bool
            Flag enabling the use of penalty function, when two exponent in
            any shell are too close the objective is multiplied by a penalty
            factor calculated in :py:func:`get_penalty
            <chemtools.basisopt.get_penalty>`

        penaltykwargs : dict
            Keyword arguments for the penalty function, default
            ``{'alpha' : 25.0, 'smallestonly' : True}``
    '''

    def __init__(self, objective=None, core=None, template=None,
                 regexp=None, verbose=False, code=None, optalg=None, mol=None,
                 fsopt=None, staticbs=None, fname=None, uselogs=True,
                 runcore=False, penalize=None, penaltykwargs=None,
                 logfile=None):

        self.fsopt = fsopt
        self.staticbs = staticbs
        self.objective = objective
        self.regexp = regexp
        self.template = template
        self.verbose = verbose
        self.core = core
        self.code = code
        self.optalg = optalg
        self.mol = mol
        self.result = None
        self.fname = fname
        self.uselogs = uselogs
        self.penalize = penalize
        self.penaltykwargs = penaltykwargs
        self.logfile = logfile

        if runcore:
            self.function = run_core_energy
        else:
            self.function = run_total_energy

    @property
    def fname(self):
        return self._fname

    @fname.setter
    def fname(self, value):
        '''
        Set the name of the job/input file for the code used for single point
        calculations.
        '''
        if value is None:
            self._fname = 'bso_' + ''.join(random.choice(string.ascii_letters)
                                           for _ in range(10)) + '.inp'
        else:
            self._fname = value

    @property
    def optalg(self):
        return self._optalg

    @optalg.setter
    def optalg(self, value):
        '''
        set default optimization options if opts not given
        '''

        # defaults
        tol = 1.0e-4
        jac = False
        mxi = 100

        if value is None:
            self._optalg = {"method": "Nelder-Mead",
                            "tol": tol,
                            "jacob": None,
                            "options": {"maxiter": 100,
                                        "disp": True,
                                        }
                            }

        elif isinstance(value, dict):
            if value.get('method').lower() == "nelder-mead":
                value['jacob'] = None
            else:
                value['jacob'] = jac
            if 'tol' not in value.keys():
                value['tol'] = tol
            if 'maxiter' not in value['options'].keys():
                value['options']['maxiter'] = mxi
            self._optalg = value
        else:
            raise ValueError("optalg should be a <dict>, got: {}".format(type(value)))

    @property
    def fsopt(self):
        return self._fsopt

    @fsopt.setter
    def fsopt(self, value):

        seqs = ["et", "even", "eventemp", "even tempered",
                "wt", "well", "welltemp", "well tempered",
                "le", "legendre", "exp", "exponents"]

        if value is None:
            raise ValueError("<fsopt> cannot be None")
        else:
            if isinstance(value, dict):
                for symbol, funlist in value.items():
                    for funtuple in funlist:
                        if funtuple[0].lower() not in ORBITALS:
                            msg = '<{}>: <shell> '.format(symbol) +\
                                  ' should be one of {}'.format(', '.join(ORBITALS)) +\
                                  ', got: {}'.format(funtuple[0])
                            raise ValueError(msg)
                        if funtuple[1] not in seqs:
                            msg = '<{}>: <seq> '.format(symbol) +\
                                  ' should be one of {}'.format(', '.join(seqs)) +\
                                  ', got: {}'.format(funtuple[1])
                            raise ValueError(msg)
            else:
                ValueError('<fsopt> should be a dict, got: {}'.format(type(value)))

            self._fsopt = OrderedDict(sorted(value.items(),
                                             key=lambda t: t[0]))

    @property
    def penaltykwargs(self):
        return self._penaltykwargs

    @penaltykwargs.setter
    def penaltykwargs(self, value):

        if value is None:
            self._penaltykwargs = {'alpha': 25.0, 'smallestonly': True}
        else:
            self._penaltykwargs = value

    @property
    def logfile(self):
        return self._logfile

    @logfile.setter
    def logfile(self, value):
        '''
        if a filename is specified open a text file for writing with
        line buffering
        '''

        if value is None:
            self._logfile = None
            self.log = sys.stdout
        else:
            self._logfile = value
            self.log = open(value, 'wt', buffering=1)
            atexit.register(self.finalize)

    def finalize(self):

        if self.log is not sys.stdout:
            self.log.close()

    def header(self):
        '''
        Return the basic information about the execution environment and
        optimization settings.
        '''

        script = sys.argv[0]
        path = os.path.abspath(os.path.dirname(script))

        out = "\n".join(["Script name : {0}".format(script),
                         "Workdir     : {0}".format(path),
                         "Start time  : {0}".format(datetime.datetime.today()),
                         ])

        return out

    def jobinfo(self):
        '''Return the information on the optimization objects and parameters'''

        # for name, obj in [("code", code), ("job", job), ("mol", mol),
        #                ("bsnoopt", bsnoopt), ("bsopt", bsopt), ("opt", opt)]:
        # TODO add printing of static basis set
        attrs = ['code', 'mol', 'optalg']
        out = ''
        for attr in attrs:
            out += attr.upper().center(80, "=") + '\n'
            obj = getattr(self, attr)
            if isinstance(obj, dict):
                out += pprint.pformat(obj)
            else:
                out += str(obj)
        return out

    def get_x0(self):
        '''
        Collect all the parameters in an array of consecutive elements.
        '''

        x0 = np.empty(0)
        for fperatom in self.fsopt.values():
            for fs in fperatom:
                if fs[1].lower() in ['exp', 'exponents'] and self.uselogs:
                    x0 = np.concatenate((x0, np.log(fs[-1])), axis=0)
                else:
                    x0 = np.concatenate((x0, np.array(fs[-1])), axis=0)
        return x0

    def run(self):
        '''
        Start the basis set optimization

        Returns:
            res : OptimizeResult
                An instance of the ``scipy.optimize.OptimizeResult`` class
        '''

        self.log.write(self.header())
        self.log.write("\n" + "\n".join(["=" * 80,
            "STARTING OPTIMIZATION".center(80, "="), "=" * 80]) + "\n\n")
        self.log.write(self.jobinfo())
        self.log.flush()

        starttime = time.time()

        x0 = self.get_x0()
        self.result = minimize(self.function, x0,
                               args=(self,),
                               method=self.optalg["method"],
                               jac=self.optalg['jacob'],
                               tol=self.optalg["tol"],
                               options=self.optalg["options"])
        self.log.write(str(self.result))
        self.log.write("Elapsed time : {0:>20.3f} sec".format(time.time() -
                                                              starttime))

    def get_basis(self, name=None, element=None):
        '''
        Construct the BasisSet object from the result of the
        optimized exponents and function definition.

        Args:
            name : str
                Name to be assigned to the basis set
            element : str
                Element symbol for the basis set

        Returns:
            basis : chemtools.basisset.BasisSet
                :py:class:`BasisSet <chemtools.basisset.BasisSet>` object with
                the optimized functions
        '''

        bsdict = {}
        # get number of parameters per atom
        npars = [sum(get_num_params(t) for t in funs) for funs in self.fsopt.values()]
        x0peratom = sliceinto(self.result.x, npars)
        for (atom, funs), xpars in zip(self.fsopt.items(), x0peratom):

            bsdict[atom] = BasisSet.from_optpars(xpars, funs, name=name,
                                                 element=element,
                                                 explogs=self.uselogs)

        if self.staticbs is not None:
            if isinstance(self.staticbs, dict):
                common_atoms = set(self.staticbs.keys()) & set(bsdict.keys())
                if common_atoms:
                    for atom in common_atoms:
                        bsdict[atom].append(self.staticbs[atom])
                diff_atoms = set(self.staticbs.keys()) - set(bsdict.keys())
                if diff_atoms:
                    for atom in diff_atoms:
                        bsdict[atom] = self.staticbs[atom]
            elif isinstance(self.staticbs, BasisSet):
                if self.staticbs.element in set(bsdict.keys()):
                    bsdict[atom].append(self.staticbs)
                else:
                    bsdict[self.staticbs.element] = self.staticbs

        return bsdict


def get_basis_dict(bso, x0):
    '''
    Return a dictionary with :py:class:`BasisSet` objects as values and
    element symbols as keys. The dictionary is composed based on the current
    parameters ``x0`` and attributes of the :py:class:`BSOptimizer` including
    the ``staticbs``
    '''

    bsdict = dict()
    for atom, functs in bso.fsopt.items():
        bsdict[atom] = BasisSet.from_optpars(x0, funs=functs, name='opt',
                                             element=atom, explogs=bso.uselogs)

    if bso.staticbs is not None:
        if isinstance(bso.staticbs, dict):
            common_atoms = set(bso.staticbs.keys()) & set(bsdict.keys())
            if common_atoms:
                for atom in common_atoms:
                    bsdict[atom].append(bso.staticbs[atom])
            diff_atoms = set(bso.staticbs.keys()) - set(bsdict.keys())
            if diff_atoms:
                for atom in diff_atoms:
                    bsdict[atom] = bso.staticbs[atom]
        elif isinstance(bso.staticbs, BasisSet):
            if bso.staticbs.element in set(bsdict.keys()):
                bsdict[atom].append(bso.staticbs)
            else:
                bsdict[bso.staticbs.element] = bso.staticbs

    return bsdict


def get_penalty(bsdict, alpha=25.0, smallestonly=True):
    '''
    For a given dict of basis sets calculate the penalty for pairs of
    exponents being too close together.

    Args:
        bsdict : dict
            Dictionary of :py:class:`BasisSet` objects

        alpha : float
            Parameter controlling the magnitude and range of the penalty

        smallestonly : bool
            A flag to mark whether to use only the smallest ratio to calculate
            the penalty or all smallest ratios from each shell and calculate
            the penalty as a product of individual penalties

    For each basis and shell within the basis ratios between pairs of sorted
    exponents are calculated. The minimal ratio (closest to 1.0) is taken to
    calculate the penalty according to the formula

    .. math::

       P = 1 + \\exp(-\\alpha (r - 1))

    where :math:`r` is the ratio between the two closest exponents (>1).
    '''

    minratios = []
    for basis in bsdict.values():
        for functions in basis.functions.values():
            exps = np.sort(functions['e'])
            ratios = exps[1:] / exps[:-1]
            minratios.append(np.min(ratios))

    if smallestonly:
        return 1 + np.exp(-alpha * (np.min(np.array(minratios)) - 1))
    else:
        return np.prod(1 + np.exp(-alpha * (np.array(minratios) - 1)))


def run_total_energy(x0, *args):
    '''
    Funtion for running a single point calculation and parsing the resulting
    energy (or property) as specified by the objective function.

    Args:
        x0 : list or numpy.array
            contains a list of parameters to be optimized, may be
            explicit exponents or parametrized exponents in terms
            of some polynomial
        args : tuple of dicts
            bsopt, bsnoopt, code, job, mol, opt, needed for writing
            input and parsing output

    Returns:
        parsed result of the single point calculation as speficied by the
        objective function in the "job" dictionary

    '''

    # unpack the args tuple for code readability
    bso = args[0]

    for atom, functs in bso.fsopt.items():
        ni = 0
        nt = 0
        for shell, seq, nf, params in functs:
            nt += nf
            if seq not in ['le', 'legendre'] and not bso.uselogs:
                x0[ni:nt] = np.abs(x0[ni:nt])
            ni += nf

    bsdict = get_basis_dict(bso, x0)

    # set the penalty value
    if bso.penalize:
        penalty = get_penalty(bsdict, **bso.penaltykwargs)
    else:
        penalty = 1.0

    if bso.verbose:
        bso.log.write("Current exponents being optimized:\n")
        for atom, functs in bso.fsopt.items():
            basis = BasisSet.from_optpars(x0, funs=functs, name='opt',
                                          element=atom, explogs=bso.uselogs)
            bso.log.write(atom + basis.print_functions())
            bso.log.flush()

    bso.code.write_input(fname=bso.fname, template=bso.template, basis=bsdict,
                         mol=bso.mol, core=bso.core)
    output = bso.code.run(bso.fname)
    if bso.code.accomplished(output):

        if callable(bso.objective):
            objective = bso.objective(output)
        else:
            objective = bso.code.parse(output, bso.objective, bso.regexp)

        if objective is None:
            raise ValueError("Unable to parse the objective, check output")
        if bso.verbose:
            bso.log.write("{0:<s}".format("Job Terminated without errors\n"))
            bso.log.write("x0 : " + ", ".join([str(x) for x in x0]) + "\n")
            bso.log.write("\n{0:<20s} : {1:>30s}\n".format("Output", output))
            bso.log.write("{0:<20s} : {1:>30.10f}\n".format(str(bso.objective), objective))
            bso.log.write("{0:<20s} : {1:>30.10f}\n".format("Objective", objective * penalty))
            bso.log.write("=" * 80 + "\n")
            bso.log.flush()
        return objective * penalty
    else:
        raise ValueError("something went wrong, check output {0:s}".format(output))


def run_core_energy(x0, *args):
    '''
    Funtion for running two single point calculations and parsing the resulting
    energy (or property) as specified by the objective function, primarily
    designed to extract core energy.

    Args:
        x0: list or numpy.array
            contains a list of parameters to be optimized, may be
            explicit exponents or parametrized exponents in terms
            of some polynomial
        args: tuple of dicts
            bsopt, bsnoopt, code, job, mol, opt, needed for writing
            input and parsing output
    Returns:
        parsed result of the single point calculation as speficied by the
        objective function in the "job" dictionary
    '''

    # unpack the args tuple for code readability
    bso = args[0]

    for atom, functs in bso.fsopt.items():
        ni = 0
        nt = 0
        for shell, seq, nf, params in functs:
            nt += nf
            if seq not in ['le', 'legendre'] and not bso.uselogs:
                x0[ni:nt] = np.abs(x0[ni:nt])
            ni += nf

    bsdict = get_basis_dict(bso, x0)

    # set the penalty value
    if bso.penalize:
        penalty = get_penalty(bsdict, bso.penaltykwargs)
    else:
        penalty = 1.0

    if bso.verbose:
        bso.log.write("Current exponents being optimized:\n")
        for atom, functs in bso.fsopt.items():
            basis = BasisSet.from_optpars(x0, funs=functs, name='opt',
                                          element=atom, explogs=bso.uselogs)
            bso.log.write(atom + "\n" + basis.print_functions())
            bso.log.flush()

    citote = []
    stats = []
    base = os.path.splitext(bso.fname)[0]
    inputs = [base + "_core" + str(sum(x)) + ".inp" for x in bso.core]

    for inpname, core in zip(inputs, bso.core):
        bso.code.write_input(fname=inpname, core=core, basis=bsdict,
                             mol=bso.mol, template=bso.template)

    outputs = bso.code.run_multiple(inputs)
    for output in outputs:
        citote.append(bso.code.parse(output, bso.objective, bso.regexp))
        stats.append(bso.code.accomplished(output))

    if stats[0] and stats[1]:
        if bso.verbose:
            bso.log.write("x0 : " + ", ".join([str(x) for x in x0]) + "\n")
            bso.log.write("{0:<20s} : {1:>30s} {2:>30s}\n".format("Terminated OK",
                                                                str(stats[0]),
                                                                str(stats[1])))
            bso.log.write("{0:<20s} : {1:>30.10f} {2:>30.10f}\n".format(str(bso.objective),
                                                              citote[0],
                                                              citote[1]))
            bso.log.write("-" * 84)
        coreenergy = citote[0] - citote[1]
        if coreenergy > 0.0:
            coreenergy = -1.0 * coreenergy
        if bso.verbose:
            bso.log.write("{0:<20s} : {1:>30.10f}\n".format("Core energy",
                                                            coreenergy))
            bso.log.write("{0:<20s} : {1:>30.10f}\n".format("Objective",
                                                            coreenergy * penalty))
            bso.log.write("=" * 84)
        bso.log.flush()
        return coreenergy * penalty
    else:
        raise ValueError("something went wrong, check outputs {0:s}".format(", ".join(outputs)))


def opt_shell_by_nf(shell=None, nfs=None, max_params=5, opt_tol=1.0e-4,
                    save=False, bsopt=None, **kwargs):
    '''
    For a given shell optimize the functions until the convergence criterion
    is reached the energy difference for two consecutive function numbers is
    less than the threshold

    Kwargs:
        shell : string
            string label for the shell to be optimized
        nfs : list of ints
            list of integers representing the number of basis functions to be
            inceremented in the optimization,
        max_params : int
            maximal number of parameters to be used in the legendre expansion,
            (length of the expansion)
        opt_tol : float
            threshold controlling the termination of the shell optimization,
            if energy difference between two basis sets with subsequent number
            of functionsis larger than this threshold, another function is
            added to this shell and parameters are reoptimized,
        save : bool
            a flag to trigger saving all the optimized basis functions for each
            shell,
        kwargs : dict
            options for the basis set optimization driver, see driver function
            from the basisopt module

    Returns:
        BasisSet object instance with optimized functions for the specified
        shell

    Raises:
        ValueError:
            when `shell` is different from `s`, `p`, `d`, `f`, `g`, `h`, `i`
            when number of parameters equals 1 and there are more functions
            when there are more parameters than functions to optimize
    '''

    _shells = ['s', 'p', 'd', 'f', 'g', 'h', 'i']

    if shell not in _shells:
        raise ValueError('shell must be one of the following: {}'.format(", ".join(_shells)))

    if len(bsopt['params'][0]) == 1 and min(nfs) != 1:
        raise ValueError("1 parameter and {0} functions doesn't make sense".format(min(nfs)))

    if min(nfs) < len(bsopt['params'][0]):
        raise ValueError("more parameters ({0}) than functions to optimize ({1})".format(len(bsopt['params'][0]), min(nfs)))

    bsopt["typ"] = "legendre"
    e_last = 0.0
    x_last = bsopt['params'][0]

    for nf in nfs:
        bsopt['nfpshell'] = [0]*_shells.index(shell) + [nf]
        res = driver(bsopt=bsopt, **kwargs)

        print("Completed optimization for {0:d} {1}-type functions".format(nf, shell))

        print("{s:<25s} : {v:>20.10f}".format(s="Current Function value", v=res.fun))
        print("{s:<25s} : {v:>20.10f}".format(s="Previous Function value", v=e_last))
        print("{s:<25s} : {v:>20.10f}".format(s="Difference", v=abs(res.fun-e_last)))

        if abs(res.fun - e_last) < opt_tol:
            print("Basis saturated with respect to threshold")
            bsopt['nfpshell'] = nfps_last
            if save:
                save_basis(x_last, bsopt)
            return BasisSet.from_optdict(x_last, bsopt)
        else:
            print("Threshold exceeded, continuing optimization.")
            x_last = res.x.tolist()
            nfps_last = bsopt['nfpshell']
            if len(bsopt['params'][0]) < max_params:
                print("adding more parameters")
                restup = tuple(res.x)
                # this assumption should be revised, improved or justified
                # suggestion: maybe change to average of existing parameters
                restup += tuple([abs(min(restup)*0.25)])
                bsopt['params'] = [restup,]
            else:
                print("not adding more parameters")
                bsopt['params'] = [tuple(res.x),]
            e_last = res.fun
    else:
        print("Supplied number of functions exhausted but the required accuracy was not reached")
        if save:
            save_basis(res.x.tolist(), bsopt)
        return BasisSet.from_optdict(res.x.tolist(), bsopt)


def opt_multishell(shells=None, nfps=None, guesses=None, max_params=5, opt_tol=1.0e-4, save=False, bsopt=None, **kwargs):
    '''
    Optimize a basis set by saturating the function space shell by shell

    Kwargs:
        shells (list of strings):
            list of shells to be optimized, in the order the optimization should
            be performed,

        nfps (list of lists of integers):
            list specifying a set of function numbers to be scanned per each
            shell,

        guesses (list of lists of floats):
            list specifying a set of starting parameters per each shell,

        max_params (int)
            maximal number of parameters to be used in the legendre expansion,
            (length of the expansion)

        opt_tol (float):
            threshold controlling the termination of the shell optimization,
            if energy difference between two basis sets with subsequent number
            of functionsis larger than this threshold, another function is
            added to this shell and parameters are reoptimized

        kwargs:
            options for the basis set optimization driver, see driver function
            from the basisopt module
    '''

    # bsnoopt needs to exists since it will be appended with optimized
    # functions after each shell is optimized
    if kwargs['bsnoopt'] is None:
        kwargs['bsnoopt'] = BasisSet.from_dict({"element"   : "Be",
                                      "functions" : {}})

    # begin the main loop over shells, nr f per shell and guesses
    for shell, nfs, guess in zip(shells, nfps, guesses):

        header = " Beginning optimization for {s:s} shell ".format(s=shell)
        print("="*100)
        print(format(header, '-^100'))
        print("="*100)

        bsopt['params'] = [tuple(guess)]
        # begin the optimization for a given shell and store the optimized
        # functions under optimized_shell
        optimized_shell = opt_shell_by_nf(shell, nfs, max_params=max_params,
                                          opt_tol=opt_tol, save=save,
                                          bsopt=bsopt, **kwargs)
        # add the optimized shell to the total basis set
        kwargs['bsnoopt'].add(optimized_shell)

    # save the final complete basis set
    basis_dict = vars(kwargs['bsnoopt'])
    with open("final.bas", 'wb') as ff:
        ff.write(str(basis_dict))



# -*- coding: utf-8 -*-

#The MIT License (MIT)
#
#Copyright (c) 2014 Lukasz Mentel
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import argparse
import os
import socket
import subprocess


def main():
    '''
    Script for submitting gamessus jobs to the queue.
    '''

    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    parser.add_argument("input",
                        help='gamessus input file to be executed')
    parser.add_argument("-d",
                        "--dryrun",
                        action="store_true",
                        help="write the pbs script but don't submit to the queue, default-False")
    parser.add_argument("-e",
                        "--extrafiles",
                        nargs="+",
                        default=[],
                        help="additional files that need to be copied to the node/scratch")
    parser.add_argument("-g",
                        "--gmsver",
                        default="00",
                        help="version of gamess executable, default=00")
    group.add_argument("-n",
                       "--nodes",
                       default="1",
                       help="number of nodes, default=1")
    group.add_argument("-H",
                       "--HOST",
                       default="",
                       help="destination hostname, default=''")
    parser.add_argument("-p",
                        "--ppn",
                        default="1",
                        help="number of processes per node, default=1")
    parser.add_argument("-s",
                        "--usescratch",
                        action="store_true",
                        help="use node scratch directory (/scratch/$USER)",
                        default=False)
    parser.add_argument("-q",
                        "--queue",
                        default="default",
                        help="destination queue, default=default")
    parser.add_argument("-t",
                        "--walltime",
                        default="120:00:00",
                        help="walltime in the format HH:MM:SS")
    args = vars(parser.parse_args())
    submit_pbs(args)


def set_defaults(args):

    args['workdir'] = os.getcwd()
    args['home'] = os.getenv("HOME")
    if socket.gethostname() == "boron.chem.umk.pl":
        args['scratch'] = os.path.join('/scratch', os.getenv('USER'))
        args['rungms'] = '/share/apps/gamess-may2013/rungms'
    elif socket.gethostname() in ["login1.lisa.surfsara.nl",
                                  "login2.lisa.surfsara.nl"]:
        args['scratch'] = os.getenv("TMPDIR")
        args['rungms'] = 'rungms'
    args['local_scr'] = os.path.join(os.getenv('HOME'), 'scratch')
    args['jobname'] = os.path.splitext(args["input"])[0]
    args['outfile'] = args['jobname'] + ".log"
    args['errfile'] = args['jobname'] + ".err"
    args['datfile'] = args['jobname'] + ".dat"
    args['script_name'] = "run." + args['jobname']
    return args


def remove_dat(path, datfile):
    '''Remove the dat file from the ASCII scratch.'''
    if os.path.exists(os.path.join(path, datfile)):
        os.remove(os.path.join(path, datfile))


def submit_pbs(args):
    '''
    Write the run script for PBS and submit it to the queue.
    '''

    args = set_defaults(args)
    remove_dat(args["local_scr"], args["datfile"])

    with open(args['script_name'], 'w') as script:
        script.write("#PBS -S /bin/bash\n")
        if args['HOST'] != "":
            script.write("#PBS -l nodes={0}:ppn={1}\n".format(args['HOST'], args['ppn']))
        else:
            script.write("#PBS -l nodes={0}:ppn={1}\n".format(args['nodes'], args['ppn']))
        if "mem" in args.keys():
            script.write("#PBS -l mem={0}\n".format(args["mem"]))
        script.write("#PBS -l walltime={0}\n\n".format(args['walltime']))
        #script.write('export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/share/apps/lib\n\n')
        script.write("#PBS -N {}\n".format(args["jobname"]))
        if args["queue"] != "default":
            script.write("#PBS -q {}\n".format(args["queue"]))
        script.write("cd $PBS_O_WORKDIR\n")
        if args['usescratch']:
            wrkdir = os.path.join(args['scratch'], args['jobname'])
            script.write("mkdir -p {}\n".format(wrkdir))
            files = args['jobname']
            if args['extrafiles']:
                files += ' ' + ' '.join(args['extrafiles'])
            script.write('cp -t {0} {1}\n'.format(wrkdir, files))
            script.write('cd {0}\n'.format(wrkdir))
        script.write("\n{0:<s} {1} {2:<s} {3:<s} \n".format(args['rungms'],
                     args['jobname'], args['gmsver'], args['ppn']))

    # submit the job to the queue if requested
    if args['dryrun']:
        print("Created job script: {0}\n NOT submitting to the queue\nbye...".format(args['script_name']))
    else:
        print("Created job script: {0}\nsubmitting to the queue".format(args['script_name']))
        output = subprocess.check_output(["qsub", args['script_name']])
        pid = output.split(".")[0]
        return args['jobname'] + ".o" + pid


def submit_slurm(args):
    '''
    Write the run script for SLURM and submit it to the queue.
    '''

    with open(args['script_name'], 'w') as script:

        script.write("#!/bin/bash\n")
        script.write("#SBATCH -t {0:<s} \n".format(args["walltime"]))
        script.write("#SBATCH -N {0:<s} \n".format(args["nodes"]))
        script.write("#SBATCH --ntasks-per-node={0:<s} \n".format(args["ppn"]))
        script.write("#SBATCH -C {0:<s} \n".format(args["nodeType"]))
        script.write("#SBATCH -o {0:>s}.%J \n".format(args['outfile']))
        script.write("#SBATCH -e {0:>s}.%J \n".format(args['errfile']))

        if int(args["walltime"].split(':')[0])*60 + int(args["walltime"].split(':')[1]) > 60 and args["nodeType"] == "thin":
            script.write("#SBATCH -p normal \n\n")
        elif int(args["walltime"].split(':')[0])*60 + int(args["walltime"].split(':')[1]) < 60 and args["nodeType"] == "thin":
            script.write("#SBATCH -p short \n\n")
        elif args["nodeType"] == "fat":
            script.write("#SBATCH -p fat \n\n")

        if args["mail"] != "":
            script.write("#SBATCH --mail-type=ALL --mail-user={0:<s}\n\n".format(args["mail"]))
        script.write("mkdir -p $TMPDIR/{0}\n".format(args['jobname']))
        script.write("cp {0} $TMPDIR/{1} \n".format(args["input"], args['jobname']))
        script.write("cd $TMPDIR/{0}\n\n".format(args['jobname']))

        script.write("module load gamess-us \n")
        script.write("{0} {1} \n".format(args['rungms'], args['jobname']))
        #script.write("cp {0} {1}\n".format(output, workdir))

    # submit the job to the queue if requested
    if args["dryrun"]:
        print("Created job script: {0}\n NOT submitting to the queue\nbye...".format(args['script_name']))
    else:
        print("Created job script: {0}\nsubmitting to the queue".format(args['script_name']))
        sublog = open(args['jobname'] + ".sublog", 'w')
        proc = subprocess.Popen(["sbatch", args['script_name']], stdout=sublog, stderr=sublog)
        sublog.close()


def submit_ll(args):
    '''
    Write the run script for LoadLeveller and submit it to the queue.
    '''

    with open(args['script_name'], 'w') as script:

        if args['ppn'] > 1 or args['nodes'] > 1:
            job_type = "parallel"
            script.write("# @ node = {0}\n".format(args['nodes']))
            script.write("# @ tasks_per_node = {0}\n".format(args['ppn']))
            script.write("# @ network.MPI_LAPI = sn_all,not_shared,US \n")
        else:
            job_type = "serial"
            script.write("# @ node_usage = shared \n")

        script.write("#\n# @ notification = never\n#\n# @ input = /dev/null\n")
        script.write("# @ output = {0}.$(jobid)\n".format(args['outfile']))
        script.write("# @ error = {0}.$(jobid)\n".format(args['errfile']))
        script.write("# @ wall_clock_limit = {0}\n#\n".format(args['walltime']))
        script.write("# @ job_type = {0}\n#\n".format(job_type))
        script.write("# @ queue \n#\n")
        script.write("mkdir -p {0} \n".format(args['scratch']))
        script.write("cp {0} {1} \n".format(
            os.path.join(args['workdir'], args['jobname']),
            os.path.join(args['scratch'], args['jobname'])))
        script.write("cd {0} \n".format(args['scratch']))
        script.write("module load gamess-us/May2012 \n")
        script.write("{0} {1} \n".format(args['rungms'], args['jobname']))

    # submit the job to the queue if requested
    if args["dryrun"]:
        print("Created job script: {0}\n NOT submitting to the queue\nbye...".format(args['script_name']))
    else:
        print("Created job script: {0}\nsubmitting to the queue".format(args['script_name']))
        sublog = open(args['jobname'] + ".sublog", 'w')
        proc = subprocess.Popen(["llsubmit", args['script_name']], stdout=sublog, stderr=sublog)
        sublog.close()


if __name__ == "__main__":
    main()
